#pylint: disable=too-many-locals

"""
The :py:mod:`paramsregression` module uses neural net classifiers to predict parameters
for error handling workflows.

.. Note::

   Regression part is a work in progress. Need more data.

:author: Daniel Abercrombie <dabercro@mit.edu>
"""

from __future__ import print_function

import sys
import json

from sklearn.neural_network import MLPClassifier

def convert_to_dense(errors, keys=None, allerrors=None, allsites=None):
    """
    Take a dictionary of sparse matrices,
    where the sparse matrices have keys of error code and them site names,
    and return the equivalent dense matrices.

    :param dict errors: The container for sparse matrices.
    :param list keys: Keys for each of the matrices.
                      Defaults to `'good_sites'` and `'bad_sites'`
    :param list allerrors: An ordered list of all the errors to consider
                           If both this and `allsites` are blank,
                           this function will just pull lists from the sparse matrices
    :param list allsites: An ordered list of all the sites to consider
    :returns: Container for two dense matrices
    :rtype: dict of lists of lists
    """
    keys = keys or ['good_sites', 'bad_sites']

    if not allerrors and not allsites:
        # First get all errors and all sites
        allerrors = set()
        allsites = set()

        for status in keys:
            for error, sites in errors[status].items():
                allerrors.add(int(error))
                for site in sites:
                    allsites.add(site)

        allerrors = sorted(allerrors)
        allsites = sorted(allsites)

    # Build the dense output
    output = {}
    for status in keys:
        output[status] = [[0] * len(allsites) for _ in allerrors]
        for i_error, error in enumerate(allerrors):
            for i_site, site in enumerate(allsites):
                output[status][i_error][i_site] += errors[status].get(str(error), {}).get(site, 0)

    return output


def get_classifier(raw_data, parameter, **kwargs):
    """
    Fit a classifier.
    If the module is run as a script,
    just print the training and test data output.
    Otherwise, return the classifier for farther use.

    :param dict raw_data: Raw data in the form of output from
                          :py:func:`actionshistorylink.dump_json`.
    :param str parameter: The parameter to classify.
    :param kwargs: These are kwargs for the ``sklearn.neural_network.MLPClassifier``
                   that is running underneath.
    :returns: Trained classifier model
    :rtype: sklearn.neural_network.MLPClassifier
    """

    primary_ids = sorted({key.split('/')[1] for key in raw_data.keys()})

    # Only split samples when running interactive tests
    training_ids = primary_ids[0::2] if __name__ == '__main__' else primary_ids

    training_data = []
    training_target = []
    testing_data = []
    testing_target = []

    class_labels = []

    # Prepare the data

    allerrors = set()
    allsites = set()

    for key in sorted(raw_data):
        for status in ['good_sites', 'bad_sites']:
            matrix = raw_data[key]['errors'][status]
            # Only do this for sparse matrices
            if not isinstance(matrix, list):
                for error, sites in matrix.items():
                    allerrors.add(int(error))
                    for site in sites:
                        allsites.add(site)

    allerrors = sorted(allerrors)
    allsites = sorted(allsites)

    for key in sorted(raw_data):
        if key.split('/')[1] in training_ids:
            data = training_data
            target = training_target
        else:
            data = testing_data
            target = testing_target

        errors = raw_data[key]['errors']
        if not isinstance(errors['good_sites'], list):
            errors = convert_to_dense(errors, allerrors=allerrors, allsites=allsites)

        data.append(sum(errors['good_sites'] + errors['bad_sites'], []))

        param = raw_data[key]['parameters'].get(parameter, '')
        if param in class_labels:
            target.append(class_labels.index(param))
        else:
            target.append(len(class_labels))
            class_labels.append(param)

    classifier = MLPClassifier(**kwargs)
    classifier.fit(training_data, training_target)

    if __name__ == '__main__':
        # Only does the following if running an interactive test
        def print_results(data, target):
            """Print the results of predictions.

            :param list data: Errors in the format of a matrix
            :param list target: The values that the data should correspond to
            """

            output = classifier.predict(data)

            right = 0

            for want, result in zip(target, output):

                if want == result:
                    status = 'RIGHT'
                    right += 1
                else:
                    status = 'WRONG'
                print('[%s] %i : %i -- %s : %s' %
                      (status, want, result, class_labels[want], class_labels[result]))

            print('%f (%i/%i)' % (100.0 * right/len(target), right, len(target)))

        print('\nTraining:\n')
        print_results(training_data, training_target)

        print('\nTesting:\n')
        print_results(testing_data, testing_target)

    return classifier


def main():
    """This is for testing."""
    if len(sys.argv) > 2:
        parameter = sys.argv[2]
    else:
        parameter = 'action'

    with open(sys.argv[1], 'r') as input_file:
        raw_data = json.load(input_file)

    get_classifier(raw_data, parameter,
                   solver='lbfgs', hidden_layer_sizes=(100, 10))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
from __future__ import print_function
import time
import json
import smtplib
from email.mime.text import MIMEText


def onFailureRate(doc, thres=0.5):
    """
    check a workflow (represented by `doc`), if its failureRate is larger than
    0.5 AND running time is > 2days, then ALERT.

    :param doc dict: information describe a workflow
    :param thres float: threshold
    :returns: (judge result, short msg if failed)
    """

    res = (False, '')
    if doc['status'] != 'running-open': return res
    if doc['failureRate'] < thres: return res
    runningOpen = [
        tr for tr in doc['transitions'] if tr['Status'] == 'running-open'
    ]
    if not runningOpen: return res

    runningOpenTime = runningOpen[0]['UpdateTime']
    if time.time() - runningOpenTime < 2 * 24 * 60 * 60: return res

    failMsg = 'FailureRate ({}) larger than threshold({}), while running time over 2 days (started at {})'.format(
        doc['failureRate'], thres, time.ctime(runningOpenTime))

    return (True, failMsg)


AlertDefs = [
    onFailureRate,
]


def alertWithEmail(docs, recipients):
    """
    handling docs with alert emails.


    :param docs list: list of documents
    :param recipients list: list of recipients email addresses
    """

    sender = 'toolsandint-workflowmonitalert@cern.ch'

    for doc in docs:
        alertResults = [ad(doc) for ad in AlertDefs]
        positiveRes = [r for r in alertResults if r[0]]
        if positiveRes:
            shortAlertMsgs = [x[1] for x in positiveRes]
            _contentMsg = '\n\n'.join([
                '*** THIS IS A GENERATED MESSAGE, PLEASE DO NOT REPLY ***',
                'Workflow: {}'.format(doc['name']),
                'Short Summary:\n{}'.format('\n'.join([
                    '- {}'.format(s) for s in shortAlertMsgs
                ])),
                '-'* 79,
                'Full document:\n{}'.format(
                    json.dumps(
                        doc, sort_keys=True, indent=4, separators=(',', ': ')))
            ])

            contentMsg = MIMEText(_contentMsg)
            contentMsg['Subject'] = '[workflowmonit] Alert on * {} *'.format(
                doc['name'])
            contentMsg['From'] = sender
            contentMsg['To'] = ', '.join(recipients)
            s = smtplib.SMTP('localhost')
            s.sendmail(sender, recipients, contentMsg.as_string())
            s.quit()


def errorEmailShooter(msg, recipients):
    """
    forward the error message to recipients by emails

    :param msg str: error mesages
    :param recipients list: list of recipients email address
    """

    sender = 'toolsandint-workflowmonitalert@cern.ch'

    contentMsg = MIMEText(msg)
    contentMsg['Subject'] = 'Exception caught for workflowmonit'
    contentMsg['From'] = sender
    contentMsg['To'] = ', '.join(recipients)
    s = smtplib.SMTP('localhost')
    s.sendmail(sender, recipients, contentMsg.as_string())
    s.quit()


def main():

    import os
    testdoc = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'Logs/toSendDoc_190317-033802.json')
    docs = json.load(open(testdoc))
    print([(d['name'], d['failureRate']) for d in docs])

    alertWithEmail(docs, ['weinan.si@cern.ch', ])



if __name__ == "__main__":
    main()



#pylint: disable=missing-docstring

import os
import json
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse # pylint: disable=import-error

from cmstoolbox.webtools import get_json

from workflowwebtools import serverconfig

def open_statuses(location):
    if os.path.isfile(location):
        with open(location, 'r') as input_file:
            return json.load(input_file)

    components = urlparse.urlparse(location)
    cookie_stuff = serverconfig.config_dict()['data']

    return get_json(components.netloc, components.path,
                    use_https=True,
                    cookie_file=cookie_stuff.get('cookie_file'),
                    cookie_pem=cookie_stuff.get('cookie_pem'),
                    cookie_key=cookie_stuff.get('cookie_key'))


def get_manual_workflows(location):
    """
    :param str location: Either the file location or the URL of statuses.json
    :returns: list of workflows in manual assistance
    :rtype: list
    """

    return [workflow for workflow, statuses
            in open_statuses(location).items()
            if True in ['manual' in status for status in statuses]]



#pylint: disable=too-many-locals

"""
Tools for clustering workflows based on their errors.

The errors are clustered based on a generated vector.
The vector can be thought as lying on two different hyper-sphere shells.
One sphere is for the error codes that occur and
the other sphere is the sites where those errors occur.
The overall distance between two workflow will be the sum in quadrature
of the distances on these two hyper-spheres.

These distances are configurable in the server ``config.yml``.
Each sphere has a center radius, thickness, and number of errors to
be at the midpoint.
The equation to determine distance from the origin for a vector
of errors is the following.

.. math::

  \\mathrm{distance} = \\frac{d}{\\sqrt{2}} +
      2 w \\left(\\frac{|\\vec{v}|}{|\\vec{v}| + m} - 0.5\\right)

The following parameters are set in ``config.yml`` for the
site name and the error code hyperspheres separately.

- *m* is the 'midpoint' parameter is the number of errors at a given site
  or given error code that will place the workflow at the midpoint of the
  hypersphere shell.
- *d* is the 'distance' parameter is the cartesian distance between the midpoints
  of two different sites or error codes.
- *w* is the 'width' parameter which defines the total shell width that the workflow
  could possibly land on.

The direction is determined by the error code or site name distribution.
Note that this always points in the upper quadrant for a given coordinate.

The equation is chosen so that two workflows that have completely different
errors at the same site, and the error width is 0,
will end up with a distance that is equal to the error distance when clustering.
This way, site and errors can have different distance weights, and there
can be some separation for the number of errors in a workflow
(for non-zero width).

:author: Daniel Abercrombie <dabercro@mit.edu>
"""

import threading

import cherrypy
import numpy
import sklearn.cluster

from . import globalerrors
from . import serverconfig
from . import errorutils


def get_workflow_vectors(workflows, session=None, allmap=None):
    """
    Gets the errors for workflows in a list of numpy arrays

    :param str workflows: the workflows that vectors are returned for
    :param cherrypy.Session session: Stores the information for a session
    :param dict allmap: a globalerrors.ErrorInfo allmap to override the
                        session's allmap
    :return: a list of numpy arrays of errors for the workflow
    :rtype: list of numpy.array
    """
    curs = globalerrors.check_session(session, can_refresh=True)
    if not allmap:
        allmap = globalerrors.check_session(session).get_allmap()

    columns = ['errorcode', 'sitename']
    column_output = {}

    for column in columns:
        # Initialize with all zeros
        settings = serverconfig.config_dict()['cluster'][column]
        column_output[column] = [numpy.zeros(len(allmap[column])) for _ in workflows]

        cherrypy.log('Getting db_lock: 3')
        curs.db_lock.acquire()
        curs.curs.execute("SELECT SUM(numbererrors), {0}, stepname "
                          "FROM workflows "
                          "GROUP BY stepname, {0} "
                          "ORDER BY {0} ASC, stepname ASC;".format(column))

        numerrors, colval, stepname = curs.curs.fetchone() or (0, '', '')
        wfname = stepname.split('/')[1] if stepname else ''

        for icol, value in enumerate(allmap[column]):
            for iwkf, workflow in enumerate(workflows):
                while colval == value and workflow == wfname:
                    column_output[column][iwkf][icol] += numerrors
                    numerrors, colval, stepname = curs.curs.fetchone() or (0, '', '')
                    if stepname:
                        wfname = stepname.split('/')[1]

        cherrypy.log('releasing db_lock: 3')
        curs.db_lock.release()

        # Preprocessing here
        for output in column_output[column]:
            length = numpy.linalg.norm(output) or 1.0
            norm = (float(settings['distance'])/1.4142 +
                    2.0 * float(settings['width']) *
                    (length/(length + float(settings['midpoint'])) - 0.5))/length

            output *= float(norm)

    return [numpy.concatenate([column_output[col][iwkf] for col in columns]) \
                for iwkf, _ in enumerate(workflows)]


def get_clusterer(history_path, errors_path=''):
    """Use this function to get the clusterer of workflows

    :param str history_path: Path to the workflow historical data.
                             This can be a local file path or a URL.
    :param str errors_path: The errors for a given session to include
                            in the clustering
    :return: A dict of a clusterer that is fitted to historical data
             with its allmap. The keys are 'clusterer' and 'allmap'.
    :rtype: dict
    """

    cherrypy.log('Initializing cluster session')

    # This will be the location of our training data
    fake_session = {
        'info': globalerrors.ErrorInfo(history_path)
        }

    # If the path to additional errors is given, add that to the clustering data.
    if errors_path:
        errorutils.add_to_database(globalerrors.check_session(fake_session), errors_path)
        globalerrors.check_session(fake_session).set_all_lists()

    # Get the data by getting table for each workflow
    workflows = globalerrors.check_session(fake_session).return_workflows()

    # Fill the data
    cherrypy.log('Getting workflow vectors')
    data = get_workflow_vectors(workflows, fake_session)

    cherrypy.log('Number of datapoints to cluster: %i' % len(data))
    cherrypy.log('Fitting workflows...')

    settings = serverconfig.config_dict()['cluster']
    clusterer = sklearn.cluster.KMeans(n_clusters=settings['n_clusters'],
                                       n_init=settings['n_init'],
                                       n_jobs=-1)

    if not settings.get('skip'):
        clusterer.fit(numpy.array(data))

    cherrypy.log('Done')

    return {'clusterer': clusterer, 'allmap': fake_session['info'].get_allmap()}


def get_workflow_groups(clusterer, session=None):
    """Groups workflows together based on a fitted clusterer

    :param dict clusterer: is a dictionary with the clusterer fit with
                           historic data and the allmap to generate it.
                           This matches the output of :func:`get_clusterer`.
    :param cherrypy.Session session: Stores the information for a session
    :returns: A dictionary pointing workflows to a group
    :rtype: dict
    """

    errorinfo = globalerrors.check_session(session, can_refresh=True)

    if errorinfo.clusters:
        return errorinfo.clusters

    cherrypy.log('Fitting existing workflows.')

    workflows = globalerrors.check_session(session).return_workflows()
    vectors = get_workflow_vectors(workflows, session, clusterer['allmap'])

    predictions = clusterer['clusterer'].predict(numpy.array(vectors))

    cherrypy.log(str(predictions))

    for index, workflow in enumerate(workflows):
        errorinfo.clusters[workflow] = predictions[index]

    return errorinfo.clusters


def get_clustered_group(workflow, clusterer, session=None):
    """Get the group for a given workflow in this session

    :param str workflow: The workflow to get the group for.
    :param sklearn.cluster.KMeans clusterer: is the clusterer fit with
                                             historic data.
    :param cherrypy.Session session: Stores the information for a session
    :returns: List of other workflows in the same group
    :rtype: set
    """

    output = []

    predictions = get_workflow_groups(clusterer, session)

    group = predictions.get(workflow)

    if group is not None:
        for wkf, cluster in predictions.items():
            if cluster == group and wkf != workflow:
                output.append(wkf)

    return output


CLUSTER_LOCK = threading.Lock()
"""
Lock that should be acquired before running clustering functions in here
"""



"""
Module to manage users of WorkflowWebTools.

:author: Daniel Abercrombie <dabercro@mit.edu>
"""

import sqlite3
import random
import uuid
import urllib
import re
import os

import cherrypy

from passlib.hash import bcrypt
from cmstoolbox.emailtools import send_email

from . import serverconfig

def get_user_db():
    """Gets the users database in the local directory.

    :returns: the users connection, cursor
    :rtype: sqlite3.Connection, sqlite3.Cursor
    """

    conn = sqlite3.connect(os.path.join(
        serverconfig.config_dict()['workspace'],
        'users.db'))
    curs = conn.cursor()
    curs.execute('SELECT name FROM sqlite_master WHERE type="table" and name="users"')

    if not curs.fetchone():
        # Note that the only un-encrypted value will be the username
        # Email is kept (but hashed and salted) for password reset of a user
        curs.execute('CREATE TABLE users (username varchar(31) PRIMARY KEY, '
                     'email varchar(255) UNIQUE, password varchar(255), '
                     'validator varchar(255), isvalid integer)')

    return conn, curs


def validate_password(_, username, password):
    """Verifies users' logon attempts.
    This is called automatically by cherrypy,
    hence the unused parameter.

    :param str _: The hostname
    :param str username: The attempted username
    :param str password: The attempted password
    :returns: Whether or not the loging attempt succeeds
    :rtype: bool
    """

    # We have to lookup password by username instead of using
    # confirmation because people are likely to have the same
    # 'password'

    conn, curs = get_user_db()

    curs.execute('SELECT password, isvalid FROM users WHERE username=?',
                 (username,))

    passwords = list(curs.fetchall())
    conn.close()

    if len(passwords) != 1:
        return False

    return do_salt_hash(password) == passwords[0][0] and bool(passwords[0][1])


def confirmation(code, lookup='validator', return_curs=False):
    """Determine the user from the confirmation code and activate the account

    :param str code: the confirmation code for the user
    :param str lookup: The field to look up the username with
    :param bool return_curs: If false, closes the connection
    :returns: the user name if valid code, or '' if not.
              Followed by conn, curs if return_curs is True
    :rtype: str [, sqlite3.Connection, sqlite3.Cursor]
    """

    conn, curs = get_user_db()

    if lookup == 'email':
        value = code
    else:
        value = do_salt_hash(code)

    # Note that the sqlite ? doesn't seem to work for these hashes for some reason
    curs.execute('SELECT username FROM users WHERE {0}=\'{1}\''.\
                     format(lookup, value))

    users = list(curs.fetchall())

    if len(users) != 1:
        user = ''

    else:
        user = users[0][0]

        if lookup == 'validator':
            curs.execute('UPDATE users SET validator=?, isvalid=? WHERE username=?',
                         ('0', 1, user))
            conn.commit()

    if return_curs:
        return user, conn, curs

    conn.close()
    return user


def send_reset_email(email, url):
    """Generates a confirmation code for a given account and sends a reset email.

    :param str email: the email linked to the account
    :param str url: the url of the running instance
    """
    cherrypy.log('Trying to send email to %s for %s' % (email, url))

    user, conn, curs = confirmation(email, 'email', True)

    if user:

        cherrypy.log('User is %s. Generating email.' % user)

        validation_string = uuid.uuid4().hex
        stored_string = do_salt_hash(str(validation_string))

        confirm_link = (url + '/confirmuser?' +
                        urllib.parse.urlencode({'code': str(validation_string)}))

        reset_link = (url + '/resetpassword?' +
                      urllib.parse.urlencode({'code': str(validation_string)}))

        message_text = (
            'Hello ' + user +',\n\n'
            'A request to reset your password for this account has been submitted. '
            'If this request was created by you, please follow the following link: \n\n' +
            reset_link +
            '\n\nIf this request was not made by you, the previous link can be deactivated '
            'by folling the following: \n\n' +
            confirm_link +
            '\n\nand please report the incident to your local webmaster.'
            '\n\nThank you, \n'
            'Some machine'
            )

        send_email(serverconfig.config_dict()['webmaster']['email'], email,
                   'Reset account on WorkflowWebTools Instance',
                   message_text)

        cherrypy.log('Email sent.')

        curs.execute('UPDATE users SET validator=?, isvalid=? WHERE username=?',
                     (stored_string, 0, user))
        conn.commit()

    conn.close()


def resetpassword(code, password):
    """Resets the password for a user.

    :param str code: is the validation code for the account
    :param str password: is the new password
    :returns: user from the password reset
    :rtype: str
    """

    user, conn, curs = confirmation(code, return_curs=True)

    if user:
        curs.execute('UPDATE users SET password=? WHERE username=?',
                     (do_salt_hash(password), user))
        conn.commit()

    conn.close()
    return user


def add_user(email, username, password, url):
    """
    Adds the user to the users database and sends a verification email,
    if the parameters are valid.

    :param str email: The user email to send verification to.
                      Make sure to check for valid domains.
    :param str username: The username of the account to add
    :param str password: The password to be stored
    :param str url: The base url to redirect the user to for validating their account
    :returns: Success code. 0 for adding user, 1 for not adding
    :rtype: int
    """

    # Need to check parameters again so that some sneaky guy doesn't pass the javascript

    good_email = False

    for valid_email in serverconfig.get_valid_emails():
        if valid_email[0] == '@':
            if email.endswith(valid_email):
                good_email = True
                break

        elif email == valid_email:
            good_email = True
            break

    if not (good_email and re.match(r'^[A-za-z0-9]+$', username)) or password == '':
        return 1

    password = do_salt_hash(password)

    validation_string = uuid.uuid4().hex
    stored_string = do_salt_hash(str(validation_string))

    confirm_link = (url + '/confirmuser?' +
                    urllib.parse.urlencode({'code': str(validation_string)}))

    wm_email = serverconfig.config_dict()['webmaster']['email']

    store_email = do_salt_hash(email)

    conn, curs = get_user_db()

    try:
        curs.execute('INSERT INTO users VALUES (?,?,?,?,?)',
                     (username, store_email, password, stored_string, 0))
    except sqlite3.IntegrityError:
        return 1

    conn.commit()
    conn.close()

    message_text = (
        'Hello ' + username +',\n\n'
        'An account using this email has been registered on an instance of WorkflowWebTools. '
        'If this request was created by you, please follow the following link: \n\n' +
        confirm_link +
        '\n\nThank you, \n' +
        serverconfig.config_dict()['webmaster']['name']
        )

    send_email(wm_email, [email, wm_email],
               'Verify Account on WorkflowWebTools Instance',
               message_text)

    return 0


def do_salt_hash(to_hash):
    """Salt and hash an object into a storable form.

    :param str to_hash: Alternate salt and hashing to get stored variable
    :returns: A salty hash
    :rtype: str
    """

    random.seed(to_hash, version=1)

    # https://bugs.python.org/issue27742
    randint_compat = lambda lo, hi: lo + int(random.random() * (hi + 1 - lo))

    with open('keys/salt.txt', 'r') as salt_file:
        salts = [line.strip() for line in salt_file.readlines()]

    salt_len = len(salts)

    for _ in range(randint_compat(10, 30)):

        to_hash = bcrypt.encrypt(to_hash, rounds=randint_compat(5, 10),
                                 salt=salts[randint_compat(0, 200) % salt_len])
        random.seed(to_hash, version=1)

    return to_hash



#pylint: disable=too-many-locals, global-statement

"""
Generates the content for the errors pages

:author: Daniel Abercrombie <dabercro@mit.edu>
"""

import os
import sqlite3
import time
import threading

from collections import defaultdict

import cherrypy

from cmstoolbox import sitereadiness

from . import workflowinfo
from . import errorutils
from . import serverconfig
from .reasonsmanip import reasons_list

class ErrorInfo: # pylint: disable=too-many-instance-attributes
    """Holds the information for any errors for a session"""

    def __init__(self, data_location=''):
        """Initialization with a setup.
        :param str data_location: Set the location of the data to read in the info
        """

        self.data_location = data_location

        # These are setup by setup()
        self.timestamp = None
        self.conn = None
        self.curs = None
        self.db_lock = threading.Lock()
        # These are setup by set_all_lists(), which is called in setup()
        self.info = None
        self.allsteps = None
        self.readiness = None
        # This is created in clusterworkflows.get_workflow_groups()
        self.clusters = {}
        # These are set in get_workflow()
        self.workflowinfos = {}
        # These are set in get_prepid()
        self.prepidinfos = {}
        # Filled by _get_step_tables
        self._step_tables = None
        # Filled by get_step_list
        self._step_list = None

        self.setup()

    def __del__(self):
        """Delete anything left over."""
        self.teardown()

    def execute(self, query, params=None):
        """
        Locks the internal database and makes the query.

        :param str query: The query, which can include '?'
        :param tuple params: The parameters to pass into the query
        :returns: The results of the query
        :rtype: list
        """

        output = []

        self.db_lock.acquire()
        curs = self.conn.cursor()
        try:
            if params:
                curs.execute(query, params)
            else:
                curs.execute(query)

            output = list(curs.fetchall())
        finally:
            self.db_lock.release()

        return output


    def setup(self):
        """Create an SQL database from the all_errors.json generated by production"""

        self.timestamp = time.time()

        if self.data_location:
            data_location = self.data_location
        else:
            data_location = serverconfig.all_errors_path()

        # Store everything into an SQL database for fast retrival

        if isinstance(data_location, str) and data_location.endswith('.db') \
                and os.path.exists(data_location):
            self.conn = sqlite3.connect(data_location, check_same_thread=False)
            curs = self.conn.cursor()
            self.curs = curs

        else:
            self.conn = sqlite3.connect(':memory:', check_same_thread=False)
            curs = self.conn.cursor()
            self.curs = curs

            errorutils.create_table(self)
            errorutils.add_to_database(self, data_location)


        self.set_all_lists()
        self.readiness = [sitereadiness.site_readiness(site) for site in self.info[3]]

        if not self.data_location:
            current_workflows = self.return_workflows()

            prep_ids = {self.get_workflow(wf).get_prep_id() for wf in current_workflows}

            other_workflows = sum([self.get_prepid(prep_id).get_workflows() \
                                       for prep_id in prep_ids], [])

            errorutils.add_to_database(self, [new for new in other_workflows \
                                                  if new not in current_workflows])
            self.set_all_lists()
            current_workflows = self.return_workflows()

            # If all ACDCs are to be shown, include the ones with zero errors like this
            if serverconfig.config_dict().get('include_all_acdcs'):
                self.allsteps.extend(['/%s/' % zero for zero in other_workflows \
                                          if zero not in current_workflows])
                self.allsteps.sort()

            self.readiness = [sitereadiness.site_readiness(site) for site in self.info[3]]

        self.connection_log('opened')

    def set_all_lists(self):
        """
        Get sets the list of all steps, sites, and errors for an ErrorInfo object.
        This should be called if data is added to the ErrorInfo cursor manually.
        """

        def get_all(column):
            """Get list of all unique entries in the database

            :param str column: is the name of the column
            :returns: a list of unique column entries
            :rtype: list
            """

            return [entry[0] for entry in \
                        self.execute('SELECT DISTINCT {0} FROM workflows'.format(column))]

        def safe_int(element):
            """A sorting algorithm that strings don't break.

            :params str element: A string that should be a number,
                                 but is taken care of in the event that it's not.
            :returns: Either the string as an integer or the string unchanged.
            :rtype: int or str
            """
            try:
                return int(element)
            except ValueError:
                return element

        allsteps = get_all('stepname')
        allsteps.sort()
        allsites = get_all('sitename')
        allsites.sort()
        allerrors = get_all('errorcode')
        allerrors.sort(key=safe_int)

        self.info = self, allsteps, allerrors, allsites

        self.allsteps = allsteps

    def teardown(self):
        """Close the database when cache expires"""
        self._step_tables = None
        self._step_list = None

        self.conn.close()
        self.connection_log('closed')

        if self.clusters:
            self.clusters = {}

    def connection_log(self, action):
        """Logs actions on the sqlite3 connection

        :param str action: is the action on the connection
        """
        if cherrypy is not None:  # This happens while everything is being deleted
            cherrypy.log('Connection {0} with timestamp {1}'.format(action, self.timestamp))

    def get_allmap(self):
        """
        :returns: A dictionary that maps 'errorcode', 'stepname', and 'sitename'
                  to the lists of all the errors, steps, or sites
        :rtype: dict
        """

        return {  # lists of elements to call for each possible row and column
            'errorcode': self.info[2],
            'stepname':  self.info[1],
            'sitename':  self.info[3]
            }

    def return_workflows(self):
        """
        :returns: the ordered list of all workflow prep IDs that need attention
        :rtype: list
        """
        wfs = list()

        last = ''

        for step in self.allsteps:
            val = step.split('/')[1]
            if val != last:
                wfs.append(val)
                last = val

        return wfs

    def get_workflow(self, workflow):
        """
        This should be used to get the workflow info so that there is no
        redundant fetching for a single session.

        :param str workflow: The prep ID for a workflow
        :returns: Cached WorkflowInfo from the ToolBox.
        :rtype: WorkflowWebTools.workflowinfo.WorkflowInfo
        """
        if not self.workflowinfos.get(workflow):
            self.workflowinfos[workflow] = workflowinfo.WorkflowInfo(workflow)

        return self.workflowinfos[workflow]

    def get_prepid(self, prep_id):
        """
        :param str prep_id: The name of the Prep ID to check cache for
        :returns: Either cached PrepIDInfo, or a new one
        :rtype: WorkflowWebTools.workflowinfo.PrepIDInfo
        """
        if not self.prepidinfos.get(prep_id):
            self.prepidinfos[prep_id] = workflowinfo.PrepIDInfo(prep_id)

        return self.prepidinfos[prep_id]

    def get_step_list(self, workflow):
        """Gets the list of steps within a workflow

        :param str workflow: Name of the workflow to gather information for
        :returns: list of steps withing the workflow
        :rtype: list
        """

        if self._step_list is None:
            self._step_list = defaultdict(list)
            cherrypy.log('Getting db_lock: 2')
            self.db_lock.acquire()
            self.curs.execute('SELECT DISTINCT(stepname) FROM workflows ORDER BY stepname')
            for tup in self.curs.fetchall():
                stepname = tup[0]
                self._step_list[stepname.split('/')[1]].append(stepname)
            cherrypy.log('Releasing db_lock: 2')
            self.db_lock.release()

        return self._step_list[workflow]

    def _get_step_tables(self):
        """Sets the internal step tables for fast fetching"""

        # The keys are stepname, then sitereadiness
        self._step_tables = defaultdict(lambda: defaultdict(list))

        for step, ready, errors, site, code in self.execute(
                """
                SELECT stepname, sitereadiness, numbererrors, sitename, errorcode FROM workflows
                ORDER BY errorcode ASC, sitename ASC
                """):
            # Append everything to 'all' to keep the order
            self._step_tables[step]['all'].append((errors, site, code))
            # Order is not as important when we are getting sparse for different readiness
            self._step_tables[step][ready].append((errors, site, code))

    def get_step_table(self, step, readymatch=None):
        """
        Get the sparse representation of the step table.
        Fetches from an internal dictionary, so faster than database access

        :param str step: The step name for the table
        :param list readymatch: The list of site readiness statuses to match
        :returns: The list used to build the step table.
                  Each element of the list is a tuple of
                  ``(number of errors, site name, exit code)``
        :rtype: list of tuples
        """

        if self._step_tables is None:
            cherrypy.log('Setting up step tables')
            self._get_step_tables()

        keys = readymatch or ['all']

        output = []
        for key in keys:
            output.extend(self._step_tables[step][key])

        return output


GLOBAL_INFO = None
GLOBAL_LOCK = threading.Lock()


def check_session(session, can_refresh=False):
    """If session is None, fills it.

    :param cherrypy.Session session: the current session
    :param bool can_refresh: tells the function if it is safe to refresh
                             and close the old database
    :returns: ErrorInfo of the session
    :rtype: ErrorInfo
    """

    if session:
        cherrypy.log('Getting global lock: 1')
        GLOBAL_LOCK.acquire()
        if session.get('lock') is None:
            session['lock'] = threading.Lock()
        cherrypy.log('Releasing global lock: 1')
        GLOBAL_LOCK.release()

        cherrypy.log('Getting session lock')
        session['lock'].acquire()

        if session.get('info') is None:
            session['info'] = ErrorInfo()
        theinfo = session.get('info')

    else:
        cherrypy.log('Getting global lock: 2')
        GLOBAL_LOCK.acquire()
        global GLOBAL_INFO
        if GLOBAL_INFO is None:
            GLOBAL_INFO = ErrorInfo()

        theinfo = GLOBAL_INFO
        cherrypy.log('Releasing global lock: 2')
        GLOBAL_LOCK.release()

    # If session ErrorInfo is old, set up another connection
    if can_refresh and theinfo.timestamp < time.time() - \
            60*serverconfig.config_dict()['refresh_period']:
        theinfo.teardown()
        theinfo.setup()

    if session:
        cherrypy.log('Releasing session lock')
        session['lock'].release()

    return theinfo


def default_errors_format():
    """
    Gives a defaultdict with the format::

      {group1: {'errors': {group1_1: {group1_1_1: errors, group1_1_2: errors}}, group2: ...}}

    :returns: A defaultdict for building errors
    :rtype: collections.defaultdict
    """

    return defaultdict(lambda: {'errors': defaultdict(lambda: defaultdict(lambda: 0)),
                                'sub': {}, 'total': 0})


def group_errors(input_errors, grouping_function, **kwargs):
    """
    Takes inputs errors with the format::

      {group1: {'errors': {group1_1: {group1_1_1: errors, group1_1_2: errors}}, group2: ...}}

    and sums the errors into a larger group.
    This second grouping is done by the output of the grouping_function.

    :param dict input_errors: The input that will be grouped
    :param grouping_function: Takes an input, which is a key of ``input_errors``
                              and groups those keys by this function output.
    :type grouping_function: function
    :param kwargs: The keyword should point to a function.
                   That keyword will be added to the dictionary of each group.
                   It's value will be the function output with the group as an argument.
    :returns: A dictionary with the same format as the input, but with groupings.
    :rtype: defaultdict
    """

    output = default_errors_format()

    for subgroup, values in input_errors.items():

        group = grouping_function(subgroup)

        # We have three variables for everything, so we can write this by hand
        # Not ideal
        for row, row_val in values['errors'].items():
            for col, numerrors in row_val.items():
                output[group]['errors'][row][col] += numerrors

        output[group]['sub'][subgroup] = values
        output[group]['total'] += values['total']

        for key, func in kwargs.items():
            output[group][key] = func(group)

    return output


def get_step_table(step, session=None, allmap=None, readymatch=None,
                   sparse=False):
    """Gathers the errors for a step into a 2-D table of ints

    :param str step: name of the step to get the table for
    :param cherrypy.Session session: Stores the information for a session
    :param dict allmap: a globalerrors.ErrorInfo allmap to override the
                        session's allmap
    :param tuple readymatch: Match the readiness statuses in this tuple, if set
    :param bool sparse: Determines whether or not a sparse matrix is returned
    :returns: A table (made of lists) of errors for the step or a sparse dictionary of entries
    :rtype: list of lists or dict of dicts of ints
    """
    info = check_session(session)
    if not allmap:
        allmap = info.get_allmap()

    contents = info.get_step_table(step, readymatch)

    if sparse:
        output = defaultdict(lambda: defaultdict(lambda: 0))

        for numbererrors, sitename, errorcode in contents:
            output[str(errorcode)][sitename] = numbererrors

        return output

    # If not sparse

    numbererrors, sitename, errorcode = contents.pop(0) if contents else (0, '', '')
    steptable = []

    for error in allmap['errorcode']:

        steprow = []

        for site in allmap['sitename']:

            if error != errorcode or site != sitename:
                steprow.append(0)
            else:
                steprow.append(numbererrors)
                numbererrors, sitename, errorcode = contents.pop(0) if contents else (0, '', '')

        steptable.append(steprow)

    return steptable


def see_workflow(workflow, session=None):
    """Gathers the error information for a single workflow

    :param str workflow: Name of the workflow to gather information for
    :param cherrypy.Session session: Stores the information for a session
    :returns: Dictionary used to generate webpage for a requested workflow
    :rtype: dict
    """

    _, _, allerrors, allsites = check_session(session).info
    steplist = check_session(session).get_step_list(workflow)

    tables = []
    # Each key is a step, and contains a list of sites to not put in the table
    skip_site = {}

    for step in steplist:
        skip_site[step] = {'sites': [], 'index': []}
        steptable = get_step_table(step, session)
        tables.append(zip(steptable, allerrors))
        for index, site in enumerate(allsites):
            if sum([row[index] for row in steptable]) == 0:
                skip_site[step]['index'].append(index)
                skip_site[step]['sites'].append(site)

    return {
        'steplist':  zip(steplist, tables),
        'allerrors': allerrors,
        'allsites':  allsites,
        'skips': skip_site,
        'reasonslist': reasons_list(),
        }


def get_row_col_names(pievar):
    """Get the column and row for the global table view, based on user input

    :param str pievar: The variable to divide the piecharts by.
    :returns: The names of the global table rows, and the table columns
    :rtype: (str, str)
    """

    pievarmap = { # for each pievar, set row and column
        'errorcode': ('stepname', 'sitename'),
        'sitename':  ('stepname', 'errorcode'),
        'stepname':  ('errorcode', 'sitename')
        }

    # Check for valid pievar and set default
    return pievarmap[pievar if pievar in pievarmap.keys() else 'errorcode']


TITLEMAP = {
    'errorcode': 'error code',
    'stepname':  'workflow',
    'sitename':  'site name',
    }
"""Dictionary that determines how a chosen pievar shows up in the pie chart titles"""


def list_matching_pievars(pievar, row, col, session=None):
    """
    Return an iterator of variables in pievar, and number of errors
    for a given rowname and colname

    :param str pievar: The variable to return an iterator of
    :param str row: Name of the row to match
    :param str col: Name of the column to match
    :param cherrypy.Session session: stores the session information
    :returns: List of tuples containing name of pievar and number of errors
    :rtype: list
    """

    curs = check_session(session, can_refresh=True)
    rowname, colname = get_row_col_names(pievar)

    output = []

    # Let's do this very carefully and stupid for now...
    for name, num in  curs.execute(('SELECT {0}, numbererrors FROM workflows '
                                    'WHERE {1}=? AND {2}=?'.
                                    format(pievar, rowname, colname)),
                                   (row, col)):
        output.append((name, num))

    return output


def get_errors(pievar, session=None):
    """
    Gets the number of errors with the format::

      {group1: {'errors': {group1_1: {group1_1_1: errors, group1_1_2: errors}}, group2: ...}}

    where each group is a different value for the variables that
    go into the row of the global errors table.
    That is, the groups will usually be the subtask list, unless ``pievar`` is ``"stepname"``.
    In that case, the grouping is by error code.

    :param str pievar: The variable that each piechart is split into.
    :param cherrypy.Session session: Stores the information for a session
    :returns: A dictionary of 2D list of errors.
    :rtype: defaultdict
    """

    rowname, colname = get_row_col_names(pievar)

    query = 'SELECT numbererrors, {0}, {1}, {2} FROM workflows ' \
        'ORDER BY {0} ASC, {1} ASC, {2} ASC;'.format(rowname, colname, pievar)

    curs = check_session(session, True)
    contents = curs.execute(query)

    output = default_errors_format()

    for numerrors, row, col, pvar in contents:
        output[row]['errors'][col][pvar] = numerrors
        output[row]['total'] += numerrors

    return output



#!/usr/bin/env python
from __future__ import print_function

import os
import sys
import time
import sqlite3
import logging
import threading
import logging.config
try:
    from Queue import Queue
except ImportError:
    from queue import Queue # pylint: disable=import-error

import yaml
from CMSMonitoring.StompAMQ import StompAMQ
import workflowmonit.workflowCollector as wc
import workflowmonit.alertingDefs as ad

CRED_FILE_PATH = os.path.join(os.path.dirname(
    os.path.abspath(__file__)), 'credential.yml')
CONFIG_FILE_PATH = os.path.join(os.path.dirname(
    os.path.abspath(__file__)), 'config.yml')
LOGDIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Logs')
LOGGING_CONFIG = os.path.join(os.path.dirname(
    os.path.abspath(__file__)), 'configLogging.yml')


class NotFinished(Exception):
    """ Task unfinished.. """


class TimeoutQueue(Queue):

    def join_with_timeout(self, timeout):
        self.all_tasks_done.acquire()
        try:
            endtime = time.time() + timeout
            while self.unfinished_tasks:
                remaining = endtime - time.time()
                if remaining <= 0.0:
                    raise NotFinished
                self.all_tasks_done.wait(remaining)
        except:
            logger.exception(
                "Unfinished tasks detected (timeout = {}s). won't wait.".format(timeout))
        finally:
            self.all_tasks_done.release()


def worker(res, q, completedWfs, minFailureRate=0., configPath=CONFIG_FILE_PATH):
    """
    Get item from queue, work on it, append result to ``res``.

    :param list res: container to hold results
    :param list completedWfs: completed workflows, to avoid re-caching
    :param TimeoutQueue q: queue
    :param float minFailureRate: minimum failure rate
    :param str configPath: path to config file
    """

    dbPath = wc.get_yamlconfig(configPath).get(
        'workflow_status_db',
        os.path.join(os.path.dirname(os.path.abspath(__file__)),
                     'workflow_status.sqlite')
    )
    DB_UPDATE_CMD = """INSERT OR REPLACE INTO workflowStatuses VALUES (?,?,?)"""

    while not q.empty():
        wf = q.get()
        if wf.workflow in completedWfs:
            continue

        try:
            failurerate = wf.get_failure_rate()
            if failurerate > minFailureRate:
                res.append(wc.populate_error_for_workflow(wf))

            toUpdate = (
                wf.workflow,
                wf._get_reqdetail().get(wf.workflow, {}).get('RequestStatus', ''),
                failurerate
            )
            if not all(toUpdate[:-1]):
                continue
            conn = sqlite3.connect(dbPath)
            with conn:
                c = conn.cursor()
                c.execute(DB_UPDATE_CMD, toUpdate)

        except:
            logger.exception(
                'workflow "{}" has trouble caching.'.format(wf.workflow))
            pass
        finally:
            q.task_done()
    return True


def getCompletedWorkflowsFromDb(configPath):
    """
    Get completed workflow list from local status db (setup to avoid unnecessary caching)

    Workflows whose status ends with *archived* are removed from further caching.

    :param str configPath: location of config file
    :returns: list of workflow (str)
    :rtype: list
    """

    config = wc.get_yamlconfig(configPath)
    if not config:
        sys.exit('Config file: {} not exist, exiting..'.format(configPath))
    dbPath = config.get(
        'workflow_status_db',
        os.path.join(os.path.dirname(os.path.abspath(__file__)),
                     'workflow_status.sqlite')
    )

    DB_CREATE_CMD = """CREATE TABLE IF NOT EXISTS workflowStatuses (
        name TEXT PRIMARY KEY,
        status TEXT,
        failurerate REAL
    );"""
    DB_QUERY_CMD = """SELECT * FROM workflowStatuses WHERE status LIKE '%archived'"""

    res = []
    conn = sqlite3.connect(dbPath)
    with conn:
        c = conn.cursor()
        c.execute(DB_CREATE_CMD)
        for row in c.execute(DB_QUERY_CMD):
            res.append(row[0])

    return res


def updateWorkflowStatusToDb(configPath, wcErrorInfos):
    """
    update workflow status to local status db, with the information from ``wcErrorInfos``.

    :param str configPath: location of config file
    :param list wcErrorInfos: list of dicts returned by :py:func:`wc.filter_n_collect`
    :returns: True
    """

    config = wc.get_yamlconfig(configPath)
    if not config:
        sys.exit('Config path: {} not exist, exiting..'.format(configPath))
    dbPath = config.get(
        'workflow_status_db',
        os.path.join(os.path.dirname(os.path.abspath(__file__)),
                     'workflow_status.sqlite')
    )

    DB_CREATE_CMD = """CREATE TABLE IF NOT EXISTS workflowStatuses (
        name TEXT PRIMARY KEY,
        status TEXT,
        failurerate REAL
    );"""
    DB_UPDATE_CMD = """INSERT OR REPLACE INTO workflowStatuses VALUES (?,?,?)"""

    toUpdate = []
    for e in wcErrorInfos:
        entry = (
            e.get('name', ''),
            e.get('status', ''),
            e.get('failureRate', 0.)
        )
        if not all(entry[:-1]):
            continue
        toUpdate.append(entry)

    conn = sqlite3.connect(dbPath)
    with conn:
        c = conn.cursor()
        c.executemany(DB_UPDATE_CMD, toUpdate)

    return True


def buildDoc(configpath):
    """
    Given a path to the config file which contains oracle db connection info,
    returns a list of documents (each for one workflow)

    :param str configpath: location of config file
    :returns: list of documents

    :rtype: list
    """

    DB_QUERY_CMD = "SELECT NAME FROM CMS_UNIFIED_ADMIN.WORKFLOW WHERE WM_STATUS LIKE 'running%'"

    _wkfs = wc.get_workflow_from_db(configpath, DB_QUERY_CMD)
    completedWfs = getCompletedWorkflowsFromDb(configpath)
    wkfs = [w for w in _wkfs if w.workflow not in completedWfs]

    logger.info('Number of workflows to query: {}'.format(len(wkfs)))

    wc.invalidate_caches('/tmp/wsi/workflowinfo')

    results = list()

    q = TimeoutQueue(maxsize=500)
    num_threads = 500
    for i, wf in enumerate(wkfs, 1):
        q.put(wf)
        if q.full() or i==len(wkfs):
            for _ in range(num_threads):
                t = threading.Thread(target=worker, args=(results, q, completedWfs, ))
                t.daemon = True
                t.start()
            try:
                q.join_with_timeout(30*60)  # timeout 30min
            except NotFinished:
                pass

    updateWorkflowStatusToDb(configpath, results)
    logger.info('Number of updated workflows: {}'.format(len(results)))

    return results


def sendDoc(cred, docs):
    """
    Given a credential dict and documents to send, make notification.

    :param dict cred: credential required by StompAMQ
    :param list docs: documents to send
    :returns: None
    """

    if not docs:
        logger.info("No document going to be set to AMQ.")
        return []

    try:
        amq = StompAMQ(
            username = None,
            password = None,
            producer = cred['producer'],
            topic = cred['topic'],
            validation_schema = None,
            host_and_ports=[
                (cred['hostport']['host'], cred['hostport']['port'])],
            logger=logger,
            cert=cred['cert'],
            key=cred['key']
        )

        doctype = 'workflowmonit_{}'.format(cred['producer'])
        notifications = [amq.make_notification(
            payload=doc, docType=doctype)[0] for doc in docs]
        failures = amq.send(notifications)

        logger.info("{}/{} docs successfully sent to AMQ.".format(
            (len(notifications)-len(failures)), len(notifications)))
        return failures

    except Exception as e:
        logger.exception(
            "Failed to send data to StompAMQ. Error: {}".format(str(e)))
        raise


def main():

    recipients = wc.get_yamlconfig(CONFIG_FILE_PATH).get('alert_recipients', [])

    try:
        with open(LOGGING_CONFIG, 'r') as f:
            config = yaml.safe_load(f.read())
            logging.config.dictConfig(config)

        global logger
        logger = logging.getLogger('workflowmonitLogger')

        cred = wc.get_yamlconfig(CRED_FILE_PATH)
        docs = buildDoc(CONFIG_FILE_PATH)

        # handling alerts
        ad.alertWithEmail(docs, recipients)

        # backup documents
        if not os.path.isdir(LOGDIR):
            os.makedirs(LOGDIR)

        doc_bkp = os.path.join(LOGDIR, 'toSendDoc_{}'.format(
            time.strftime('%y%m%d-%H%M%S')))
        docfn = wc.save_json(docs, filename=doc_bkp, gzipped=True)
        logger.info('Document saved at: {}'.format(docfn))

        failures = sendDoc(cred=cred, docs=docs)

        failedDocs_bkp = os.path.join(
            LOGDIR, 'amqFailedMsg_{}'.format(time.strftime('%y%m%d-%H%M%S')))
        if len(failures):
            failedDocFn = wc.save_json(failures, filename=failedDocs_bkp, gzipped=True)
            logger.info('Failed message saved at: {}'.format(failedDocFn))
    except Exception as e:
        logger.exception("Exception encounted, sending emails to {}".format(str(recipients)))
        ad.errorEmailShooter(str(e), recipients)


if __name__ == "__main__":
    main()



# pylint: disable=missing-docstring, invalid-name, too-many-branches, too-many-locals

"""
A module that evaluates a model and returns the prediction
"""

from __future__ import print_function

import json
import requests

from .. import serverconfig
from ..web.templates import render


def static(workflow):
    """
    Returns the action from the static AIEH model for a given workflow
    :param str workflow: The name of the workflow to check
    :returns: The action determined by the static model
    :rtype: dict
    """

    response = json.loads(
        requests.get(serverconfig.config_dict()['aieh']['static'],
                     params={'wf': workflow}).text)

    return response


def predict(wf_obj):
    """
    Takes the errors for a workflow and makes an action prediction
    :param workflowwebtool.workflowinfo.WorkflowInfo wf_obj:
        The WorkflowInfo object that we want to perform a prediction on
    :returns: Prediction results to be passed back to a browser
    :rtype: dict
    """

    config = serverconfig.config_dict()

    if config.get('no_predict') or not config.get('aieh'):
        return {'Action': 'Suggestions turned off'}

    wf_name = wf_obj.workflow

    tasks = sorted(wf_obj.get_errors(True).keys())

    table_rows = {
        task: json.loads(
            requests.get(serverconfig.config_dict()['aieh']['models'],
                         params={'wf': wf_name, 'tsk': task}).text)
        for task in tasks}

    # Set of all models turned into a sorted list
    allmodels = table_rows[tasks[0]]

    return {
        'Action': render('predictiontable.html',
                         allmodels=allmodels,
                         tasks=tasks,
                         table_rows=table_rows)
        }



# pylint: skip-file

"""
Defines the class that runs the server
"""

# pylint: disable=no-member, no-self-use, invalid-name

# Definitely clean these
# pylint: disable=too-many-public-methods, missing-docstring, attribute-defined-outside-init


import json
import time
import datetime
import threading
import sqlite3

import cherrypy

from cmstoolbox import sitereadiness
from cmstoolbox import checkexists


from workflowwebtools import workflowinfo
from workflowwebtools import serverconfig
from workflowwebtools import manageactions
from workflowwebtools import manageusers
from workflowwebtools import reasonsmanip
from workflowwebtools import listpage
from workflowwebtools import globalerrors
from workflowwebtools import clusterworkflows
from workflowwebtools import classifyerrors
from workflowwebtools import actionshistorylink
from workflowwebtools import errorutils
from workflowwebtools.web.templates import render
from workflowwebtools.predict import evaluate

from workflowwebtools import statuses


class WorkflowTools(object):

    RESET_LOCK = threading.Lock()

    def __init__(self):
        self.lock = threading.Lock()
        self.wflock = threading.Lock()
        self.readinesslock = threading.Lock()
        self.seeworkflowlock = threading.Lock()
        self.cluster()
        self.update()

        self.markedreset = set()


    @cherrypy.expose
    def markreset(self, prepid):
        self.lock.acquire()
        self.markedreset.add(prepid)
        self.lock.release()

    def reset(self):

        self.lock.acquire()
        # Reset things is requested
        if self.markedreset:
            for pid in self.markedreset:
                for wf in self.prepids[pid].get_workflows():
                    self.wflock.acquire()
                    workflow_obj = self.workflows.pop(wf, None)
                    self.wflock.release()
                    if workflow_obj:
                        workflow_obj.reset()

                prep_obj = self.prepids.pop(pid, None)
                if prep_obj:
                    prep_obj.reset()

            self.markedreset = set()
        self.lock.release()



    def update(self):

        self.lock.acquire()
        self.workflows = {}

        try:
            for workflow in statuses.get_manual_workflows(
                    serverconfig.config_dict()['data']['all_errors']):
                self.get(workflow)

            self.prepids = {
                prepid: workflowinfo.PrepIDInfo(prepid) for prepid in
                [info.get_prep_id() for info in self.workflows.values()]
            }

            self.update_statuses()

        finally:
            self.lock.release()


    def update_statuses(self):
        coll = manageactions.get_actions_collection()
        self.site_statuses = None
        self.statuses = {
            record['workflow']: record['acted']
            for record in coll.find()
            }

    @cherrypy.expose
    def index(self):
        """
        :returns: The welcome page
        :rtype: str
        """
        return render('welcome.html')

    @cherrypy.expose
    def cluster(self):
        """
        The function is only accessible to someone with a verified account.

        Navigating to ``https://localhost:8080/cluster``
        causes the server to regenerate the clusters that it has stored.
        This is useful when the history database of past errors has been
        updated with relevant errors since the server has been started or
        this function has been called.

        :returns: a confirmation page
        :rtype: str
        """
        data = serverconfig.config_dict()['data']
        self.clusterer = clusterworkflows.get_clusterer(
            data['workflow_history'], data['all_errors'])
        return render('complete.html')

    @cherrypy.expose
    def globalerror2(self, reset=False):
        if reset:
            self.reset()
            self.update()

        return render(
            'globalerror2.html'
            )

    @cherrypy.expose
    @cherrypy.tools.json_out()
    def getprepids(self):
        return sorted(self.prepids.keys())


    @cherrypy.expose
    @cherrypy.tools.json_out()
    def predict(self, workflow):
        return evaluate.predict(self.get(workflow))

    def get_status(self, workflow):
        status = self.statuses.get(workflow)
        if status is None:
            return "none"
        return "acted" if status else "pending"

    @cherrypy.expose
    @cherrypy.tools.json_out()
    def getstatus(self, workflow):
        return {'status': self.get_status(workflow).capitalize()}

    def get(self, workflow):
        self.wflock.acquire()
        wkflow_obj = self.workflows.get(workflow)
        if wkflow_obj is None:
            self.workflows[workflow] = workflowinfo.WorkflowInfo(workflow)
            wkflow_obj = self.workflows[workflow]

        self.wflock.release()
        return wkflow_obj


    @cherrypy.expose
    @cherrypy.tools.json_out()
    def getworkflows(self, prepid):
        workflow_objs = {
            workflow: {
                'obj': self.get(workflow),
                'time': requesttime
                }
            for workflow, requesttime in
            self.prepids[prepid].get_workflows_requesttime()
            }

        workflows = [
            {"workflow": workflow,
             "status": self.get_status(workflow),
             "errors": obj['obj'].sum_errors()
            }
            for workflow, obj in workflow_objs.items()
            ]

        self.lock.acquire()
        for workflow, obj in workflow_objs.items():
            if workflow not in self.workflows:
                self.workflows[workflow] = obj['obj']
        self.lock.release()

        return sorted(
            [wkflow for wkflow in workflows if wkflow['errors']],
            key=lambda wkfl: workflow_objs[wkfl['workflow']]['time']
        )


    @cherrypy.expose
    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out()
    def submit2(self):
        input_json = cherrypy.request.json
        manageactions.submit2(cherrypy.request.login, input_json['documents'])
        self.update_statuses()
        return {'message': 'Done'}


    @cherrypy.expose
    def globalerror(self, pievar='errorcode'):
        """
        This page, located at ``https://localhost:8080/globalerror``,
        attempts to give an overall view of the errors that occurred
        in each workflow at different sites.
        The resulting view is a table of piecharts.
        The rows and columns can be adjusted to contain two of the following:

        - Workflow step name
        - Site where error occurred
        - Exit code of the error

        The third variable is used to split the pie charts.
        This variable inside the pie charts can be quickly changed
        by submitting the form in the upper left corner of the page.

        The size of the piecharts depend on the total number of errors in a given cell.
        Each cell also has a tooltip, giving the total number of errors in the piechart.
        The colors of the piecharts show the splitting based on the ``pievar``.
        Clicking on the pie chart will show the splitting explicitly
        using the :ref:`list-wfs-ref` page.

        If the steps make up the rows,
        the default view will show you the errors for each campaign.
        Clicking on the campaign name will cause the rows to expand
        to show the original workflow and all ACDCs (whether or not the ACDCs have errors).
        Following the link of the workflow will bring you to :ref:`workflow-view-ref`.
        Clicking anywhere else in the workflow box
        will cause it to expand to show errors for each step.

        :param str pievar: The variable that the pie charts are split into.
                           Valid values are:

                           - errorcode
                           - sitename
                           - stepname

        :returns: the global views of errors
        :rtype: str
        """

        # For some reasons, we occasionally have to refresh this global errors page

        errors = globalerrors.get_errors(pievar, cherrypy.session)
        if pievar != 'stepname':

            # This pulls out the timestamp from the workflow parameters
            timestamp = lambda wkf: time.mktime(
                datetime.datetime(
                    *(globalerrors.check_session(cherrypy.session).\
                          get_workflow(wkf).get_workflow_parameters()['RequestDate'])).timetuple()
                )

            errors = globalerrors.group_errors(
                globalerrors.group_errors(errors, lambda subtask: subtask.split('/')[1],
                                          timestamp=timestamp),
                lambda workflow: globalerrors.check_session(cherrypy.session).\
                    get_workflow(workflow).get_prep_id()
                )

        # Get the names of the columns
        cols = globalerrors.check_session(cherrypy.session).\
            get_allmap()[globalerrors.get_row_col_names(pievar)[1]]

        get_names = lambda x: [globalerrors.TITLEMAP[name]
                               for name in globalerrors.get_row_col_names(x)]

        template = lambda: render(
            'globalerror.html',
            errors=errors,
            decoder=json.dumps,
            columns=cols,
            pievar=pievar,
            acted_workflows=manageactions.get_acted_workflows(
                serverconfig.get_history_length()),
            readiness=globalerrors.check_session(cherrypy.session).readiness,
            get_names=get_names
            )

        try:
            return template()
        # I don't remember what kind of exception this throws...
        except Exception: # pylint: disable=broad-except
            time.sleep(1)
            return template()

    @cherrypy.expose
    @cherrypy.tools.json_out()
    def getreasons(self):
        return [
            {
                'short': shortreason,
                'long': longreason
            }
            for shortreason, longreason in
            sorted(reasonsmanip.reasons_list().items())
        ]


    @cherrypy.expose
    @cherrypy.tools.json_out()
    def workflowerrors(self, workflow):
        wkflow_obj = self.get(workflow)

        errors = wkflow_obj.get_errors(True)
        output = []

        # Need to track total sites, so we need to do nested loops here
        for step, ecs in sorted(errors.items()):
            allsites = set()

            codes = []
            for code, sites in sorted([
                    (-1 if code == 'NotReported' else int(code), sites)
                    for code, sites in ecs.items()]):

                sites = {
                    site: (num or int(code < 0))
                    for site, num in sorted(sites.items())
                }
                allsites.update(sites.keys())

                codes.append({
                    'code': code,
                    'sites': sites
                })

            output.append({
                'step': step,
                'codes': codes,
                'allsites': sorted(allsites)
            })

        return output


    @cherrypy.expose
    def seeworkflow2(self, workflow):
        return render(
            'workflowtables2.html',
            workflow=workflow
            )

    @cherrypy.expose
    def seeworkflow(self, workflow='', issuggested=''):
        """
        Located at ``https://localhost:8080/seeworkflow``,
        this shows detailed tables of errors for each step in a workflow.

        For the exit codes in each row, there is a link to view some of the output
        of the error message for jobs having the given exit code.
        This should help operators understand what the error means.

        At the top of the page, there are links back for :ref:`global-view-ref`,
        workflow logs, related JIRA tickets,
        and ReqMgr2 information about the workflow and prep ID.

        The main function of this page is to submit actions.
        Note that you will need to register in order to actually submit actions.
        See :ref:`new-user-ref` for more details.
        Depending on which action is selected, a menu will appear
        for the operator to adjust parameters for the workflows.

        Under the selection of the action and parameters, there is a button
        to show other workflows that are similar to the selected workflow,
        according to the :ref:`clustering-ref`.
        Each entry is a link to open a similar workflow view page in a new tab.
        The option to submit actions will not be on this page though
        (so that you can focus on the first workflow).
        If you think that a workflow in the cluster should have the same actions
        applied to it as the parent workflow,
        then check the box next to the workflow name before submitting the action.

        Finally, before submitting, you can submit reasons for your action selection.
        Clicking the Add Reason button will give you an additional reason field.
        Reasons submitted are stored based on the short reason you give.
        You can then select past reasons from the drop down menu to save time in the future.
        If you do not want to store your reason, do not fill in the Short Reason field.
        The long reason will be used for logging
        and communicating with the workflow requester (eventually).

        :param str workflow: is the name of the workflow to look at
        :param str issuggested: is a string to tell if the page
                                has been linked from another workflow page
        :returns: the error tables page for a given workflow
        :rtype: str
        :raises: 404 if a workflow doesn't seem to be in assistance anymore
                 Resets personal cache in the meanwhile, just in case
        """

        self.seeworkflowlock.acquire()

        output = ''

        try:
            if workflow not in \
                    globalerrors.check_session(
                            cherrypy.session, can_refresh=True).return_workflows():
                WorkflowTools.RESET_LOCK.acquire()
                info = globalerrors.check_session(cherrypy.session)
                if info:
                    info.teardown()
                    info.setup()
                WorkflowTools.RESET_LOCK.release()

                raise cherrypy.HTTPError(404)

            workflowdata = globalerrors.see_workflow(workflow, cherrypy.session)

            drain_statuses = {sitename: drain for sitename, _, drain in \
                                  sitereadiness.i_site_readiness()}

            output = render(
                'workflowtables.html',
                workflowdata=workflowdata,
                workflow=workflow,
                issuggested=issuggested,
                workflowinfo=globalerrors.check_session(cherrypy.session).get_workflow(workflow),
                readiness=globalerrors.check_session(cherrypy.session).readiness,
                drain_statuses=drain_statuses,
                last_submitted=manageactions.get_datetime_submitted(workflow)
                )
        finally:
            self.seeworkflowlock.release()

        return output


    @cherrypy.expose
    @cherrypy.tools.json_out()
    def wkfparams(self, workflow):
        """
        Get the workflow parameters
        :param str workflow: The workflow that we need parameters for
        :returns: Parameters for this workflow
        :rtype: JSON
        """

        return self.get(workflow).get_workflow_parameters()


    @cherrypy.expose
    @cherrypy.tools.json_out()
    def drainstatuses(self):
        """
        :returns: An object (dictionary) of drain statuses of sites
        :rtype: JSON
        """
        return {sitename: drain for sitename, _, drain in \
                    sitereadiness.i_site_readiness()}


    @cherrypy.expose
    @cherrypy.tools.json_out()
    def sitestatuses(self):
        """
        :returns: An object (dictionary) of drain statuses of sites
        :rtype: JSON
        """

        self.readinesslock.acquire()

        try:
            if self.site_statuses is None:
                self.site_statuses = [
                    {
                        'site': site,
                        'status': status,
                        'drain': drain
                    }
                    for site, status, drain in sitereadiness.i_site_readiness() \
                    if not site.startswith('T3')
                ]
        finally:
            self.readinesslock.release()

        return self.site_statuses


    @cherrypy.expose
    @cherrypy.tools.json_out()
    def submissionparams(self, workflow):

        wkfl_obj = self.get(workflow)
        steps = sorted(wkfl_obj.get_errors())

        return {
            'submitted': str(manageactions.get_datetime_submitted(workflow)),
            'steps': steps,
            'sitestorun': {
                step: wkfl_obj.site_to_run(step) for step in steps
            },
            'allsites': sorted(
                [{'site': site['site'], 'drain': site['drain']}
                 for site in self.sitestatuses()],
                key=lambda s: s['site']
            )
        }


    @cherrypy.expose
    @cherrypy.tools.json_out()
    def similarwfs(self, workflow):
        """
        Gives back a list of workflows that are clustered with the queried workflow.
        This is just an array of strings.

        :param str workflow: The workflow to find the group.
        :returns: List of similar workflows.
                  If the quey is not a valid workflow in the system, an empty list is returned
        :rtype: JSON
        """
        output = {'similar': [], 'acted': []}

        if serverconfig.config_dict()['cluster'].get('skip'):
            return output

        self.seeworkflowlock.acquire()

        try:
            if workflow in \
                    globalerrors.check_session(cherrypy.session,
                                               can_refresh=True).return_workflows():

                clusterworkflows.CLUSTER_LOCK.acquire()

                similar_wfs = clusterworkflows.\
                    get_clustered_group(workflow, self.clusterer, cherrypy.session)

                clusterworkflows.CLUSTER_LOCK.release()

                acted = [
                    wf for wf in manageactions.get_acted_workflows(
                        serverconfig.get_history_length()) if wf in similar_wfs
                ]

                output = {'similar': sorted(list(similar_wfs)),
                          'acted': acted}
        finally:
            self.seeworkflowlock.release()

        return output


    @cherrypy.expose
    @cherrypy.tools.json_out()
    def classifyerror(self, workflow):
        """
        :returns: An object full of informaton about the workflow errors.
                  The keys are the following:

                  - ``maxerror`` -- The error code that occurred most frequently for the workflow
                  - ``types`` -- Types of errors and exit codes
                  - ``recommended`` -- Recommended actions
                  - ``params`` -- Additional parameters to do actions

        :rtype: JSON
        """

        output = {}
        self.seeworkflowlock.acquire()

        try:
            max_error = classifyerrors.get_max_errorcode(self.get(workflow))
            main_error_class = classifyerrors.classifyerror(max_error, self.get(workflow))

            output = {
                'maxerror': max_error,
                'types': main_error_class['types'],
                'recommended': main_error_class['recommended'],
                'params': main_error_class['params_string']
            }

        finally:
            self.seeworkflowlock.release()

        return output


    @cherrypy.expose
    @cherrypy.tools.json_in()
    def update_reasons(self, reasons):
        reasonsmanip.update_reasons(reasons)
        return 'OK'


    @cherrypy.expose
    @cherrypy.tools.json_out()
    def sitesfortasks(self, **kwargs):
        """
        Accessed through a popup that allows user to submit sites for workflow
        tasks that did not have any sites to run on.
        Returns operators back the :py:func:`get_action` output.

        :param kwargs: Set up in a way that manageactions.extract_reasons_params
                       can extract the sites for each subtask.
        :returns: View of actions submitted
        :rtype: JSON
        """

        manageactions.fix_sites(**kwargs)
        return self.getaction(1)

    @cherrypy.expose
    @cherrypy.tools.json_out()
    def actionshistory(self):
        """
        This API gives the sparse matrix that can be used for training.

        :returns: History of actions on workflows
        :rtype: JSON
        """
        return actionshistorylink.dump_json()

    @cherrypy.expose
    def submitaction(self, workflows='', action='', **kwargs):
        """Submits the action to Unified and notifies the user that this happened

        :param str workflows: is a list of workflows to apply the action to
        :param str action: is the suggested action for Unified to take
        :param kwargs: can include various reasons and additional datasets
        :returns: a confirmation page
        :rtype: str
        """

        cherrypy.log('args: {0}'.format(kwargs))

        if workflows == '':
            return render('scolduser.html', workflow='')

        if action == '':
            return render('scolduser.html', workflow=workflows[0])

        output = ''

        self.seeworkflowlock.acquire()
        try:

            workflows, reasons, params = manageactions.\
                submitaction(cherrypy.request.login, workflows, action, cherrypy.session,
                             **kwargs)

            # Immediately get actions to check the sites list
            check_actions = manageactions.get_actions()
            blank_sites_subtask = []
            sites_to_run = {}
            # Loop through all workflows just submitted
            for workflow in workflows:
                # Check sites of recovered workflows
                if check_actions[workflow]['Action'] in ['acdc', 'recovery']:
                    for subtask, params in check_actions[workflow]['Parameters'].items():
                        # Empty sites are noted
                        if not params.get('sites'):
                            blank_sites_subtask.append('/%s/%s' % (workflow, subtask))
                            sites_to_run['/%s/%s' % (workflow, subtask)] = \
                                globalerrors.check_session(cherrypy.session).\
                                get_workflow(workflow).site_to_run(subtask)

            if blank_sites_subtask:
                drain_statuses = {sitename: drain for sitename, _, drain in \
                                      sitereadiness.i_site_readiness()}
                output = render('picksites.html',
                                tasks=blank_sites_subtask,
                                statuses=drain_statuses,
                                sites_to_run=sites_to_run)

            else:
                output = render('actionsubmitted.html',
                                workflows=workflows,
                                action=action,
                                reasons=reasons,
                                params=params,
                                user=cherrypy.request.login)

        finally:
            self.seeworkflowlock.release()

        return output


    @cherrypy.expose
    @cherrypy.tools.json_out()
    def getaction(self, days=0, acted=0):
        """
        The page at ``https://localhost:8080/getaction``
        returns a list of workflows to perform actions on.
        It may be useful to use this page to immediately check
        if your submission went through properly.
        This page will mostly be used by Unified though for acting on operator submissions.

        :param int days: The number of past days to check.
                         The default, 0, means to only check today.
        :param int acted: Used to determine which actions to return.
                          The following values can be used:

                          - 0 - The default value selects actions that have not been run on
                          - 1 - Selects actions reported as submitted by Unified
                          - Negative integer - Selects all actions

        :returns: JSON-formatted information containing actions to act on.
                  The top-level keys of the JSON are the workflow names.
                  Each of these keys refers to a dictionary specifying:

                  - **"Action"** - The action to take on the workflow
                  - **"Reasons"** - A list of reasons for this action
                  - **"Parameters"** - Changes to make for the resubmission
                  - **"user"** - The account name that submitted that action

        :rtype: JSON
        """
        acted = int(acted)
        if acted < 0:
            acted = None
        if acted > 1:
            acted = 1

        return manageactions.get_actions(int(days), acted=acted)

    @cherrypy.expose
    @cherrypy.tools.json_in()
    @cherrypy.tools.json_out()
    def reportaction(self):
        """
        A POST request to ``https://localhost:8080/reportaction``
        tells the WorkflowWebTools that a set of workflows has been acted on by Unified.
        The body of the POST request must include a JSON with the passphrase
        under ``"key"`` and a list of workflows under ``"workflows"``.

        An example of making this POST request is provided in the file
        ``WorkflowWebTools/test/report_action.py``,
        which relies on ``WorkflowWebTools/test/key.json``.

        :returns: A JSON summarizing results of the request. Keys and values include:

                  - `'valid_key'` - If this is false, the passphrase passed was wrong,
                                    and no action is changed
                  - `'valid_format'` - If false, the input format is unexpected
                  - `'success'` - List of actions that have been marked as acted on
                  - `'already_reported'` - List of workflows that were already removed
                  - `'does_not_exist'` - List of workflows that the database does not know

        :rtype: JSON
        """

        input_json = cherrypy.request.json

        # Is the input good?
        goodkey = (input_json['key'] == serverconfig.config_dict()['actions']['key'])
        goodformat = isinstance(input_json['workflows'], list)

        output = {'valid_key': goodkey, 'valid_format': goodformat}

        if goodkey and goodformat:
            # This output is added to by passing reference to manageactions.report_actions
            manageactions.report_actions(input_json['workflows'], output)

        return output


    @cherrypy.expose
    def explainerror(self, errorcode='0', workflowstep='/'):
        """Returns an explaination of the error code, along with a link returning to table

        :param str errorcode: The error code to display.
        :param str workflowstep: The workflow to return to from the error page.
        :returns: a page dumping the error logs
        :rtype: str
        """

        workflow = workflowstep.split('/')[1]
        if errorcode == '0' or not workflow:
            return 'Need to specify error and workflow. Follow link from workflow tables.'

        errs_explained = globalerrors.check_session(cherrypy.session).\
            get_workflow(workflow).get_explanation(errorcode, workflowstep)

        return render('explainerror.html',
                      error=errorcode,
                      explanation=errs_explained,
                      source=workflowstep)

    @cherrypy.expose
    def newuser(self, email='', username='', password=''):
        """
        New users can register at ``https://localhost:8080/newuser``.
        From this page, users can enter a username, email, and password.
        The username cannot be empty, must contain only alphanumeric characters,
        and must not already exist in the system.
        The email must match the domain names listed on the page or can
        be a specific whitelisted email.
        See :ref:`server-config-ref` for more information on setting valid emails.
        Finally, the password must also be not empty.

        If the registration process is successful, the user will recieve a confirmation
        page instructing them to check their email for a verification link.
        The user account will be activated when that link is followed,
        in order to ensure that the user possesses a valid email.

        The following parameters are sent via POST from the registration page.

        :param str email: The email of the new user
        :param str username: The username of the new user
        :param str password: The password of the new user
        :returns: a page to generate a new user or a confirmation page
        :rtype: str
        :raises: cherrypy.HTTPRedirect back to the new user page without parameters
                 if there was a problem entering the user into the database
        """

        if '' in [email, username, password]:
            return render('newuser.html',
                          emails=serverconfig.get_valid_emails())

        add = manageusers.add_user(email, username, password,
                                   cherrypy.url().split('/newuser')[0])
        if add == 0:
            return render('checkemail.html', email=email)

        raise cherrypy.HTTPRedirect('/newuser')

    @cherrypy.expose
    def confirmuser(self, code):
        """Confirms and activates an account

        :param str code: confirmation code to activate the account
        :returns: confirmation screen for the user
        :rtype: str
        :raises: A redirect the the homepage if the code is invalid
        """

        user = manageusers.confirmation(code)
        if user != '':
            return render('activated.html', user=user)
        raise cherrypy.HTTPRedirect('/')

    @cherrypy.expose
    def resetpassword(self, email='', code='', password=''):
        """
        If a user forgets his or her username or password,
        navigating to ``https://localhost:8080/resetpassword`` will
        allow them to enter their email to reset their password.
        The email will contain the username and a link to reset the password.

        This page is multifunctional, depending on which parameters are sent.
        The link actually redirects to this webpage with a secret code
        that will then allow you to submit a new password.
        The password is then submitted back here via POST.

        :param str email: The email linked to the account
        :param str code: confirmation code to activate the account
        :param str password: the new password for a given code
        :returns: a webview depending on the inputs
        :rtype: str
        :raises: 404 if both email and code are filled
        """

        if not(email or code or password):
            return render('requestreset.html')

        elif not (code or password):
            manageusers.send_reset_email(
                email, cherrypy.url().split('/resetpass')[0])
            return render('sentemail.html', email=email)

        elif not email and code:
            if not password:
                return render('newpassword.html', code=code)

            user = manageusers.resetpassword(code, password)
            return render('resetpassword.html', user=user)

        raise cherrypy.HTTPError(404)

    @cherrypy.expose
    def resetcache(self, prepid='', workflow=''):
        """
        The function is only accessible to someone with a verified account.

        Navigating to ``https://localhost:8080/resetcache``
        resets the error info for the user's session.
        It also clears out cached JSON files on the server.
        Under normal operation, this cache is only refreshed every half hour.

        :returns: a confirmation page
        :rtype: str
        """

        WorkflowTools.RESET_LOCK.acquire()
        cherrypy.log('Cache reset by: %s' % cherrypy.request.login)
        info = globalerrors.check_session(cherrypy.session)

        # Force the cache reset
        if info:
            prepids = [prepid] if prepid else info.prepidinfos.keys()
            workflows = [workflow] if workflow else \
                [wf for pid in prepids for wf in info.prepidinfos[pid].get_workflows()]

            for wkf in workflows:
                info.get_workflow(wkf).get_errors()
                info.get_workflow(wkf).reset()

            if not workflow:
                for pid in prepids:
                    info.prepidinfos[pid].reset()

            info.teardown()
            info.setup()

        WorkflowTools.RESET_LOCK.release()

        return render('complete.html')

    @cherrypy.expose
    def listpage(self, errorcode='', sitename='', workflow=''):
        """
        This returns a list of workflows, site names, or error codes
        that matches the values given for the other two variables.
        The page can be accessed directly by clicking on a corresponding pie chart
        on :ref:`global-view-ref`.

        :param int errorcode: Error to match
        :param str sitename: Site to match
        :param str workflow: The workflow to match
        :returns: Page listing workflows, site names or errors codes
        :rtype: str
        :raises: cherrypy.HTTPRedirect to 404 if all variables are filled
        """

        if errorcode and sitename and workflow:
            raise cherrypy.HTTPError(404)

        acted = [] if workflow else \
            manageactions.get_acted_workflows(
                serverconfig.get_history_length())

        # Retry after ProgrammingError
        try:
            info = listpage.listworkflows(errorcode, sitename, workflow, cherrypy.session)
        except sqlite3.ProgrammingError:
            time.sleep(5)
            return self.listpage(errorcode, sitename, workflow)

        return render('listworkflows.html',
                      workflow=workflow,
                      errorcode=errorcode,
                      sitename=sitename,
                      acted_workflows=acted,
                      info=info)

    @cherrypy.expose
    @cherrypy.tools.json_out()
    def problem_files(self, workflow):
        # Which errors do we want to check files:
        errors = [84, 85, 92, 8020, 8021, 8028]

        files_to_check = set()

        wfi = self.get(workflow)

        for error in errors:
            for file_to_add in classifyerrors.classifyerror(error, wfi)['list_of_params']:
                files_to_check.add(file_to_add)

        return sorted(files_to_check)

    @cherrypy.expose
    @cherrypy.tools.json_out()
    def file_exists(self, filename):
        output = {'xrdfs_locate': int(checkexists.exists(filename))}

        return output

    @cherrypy.expose
    @cherrypy.tools.json_out()
    def auto_acdc(self, workflow):

        action = evaluate.static(workflow)

        if action['action'] not in {'acdc'}:
            return {}

        params = self.submissionparams(workflow)

        site_map = {site['site']: site['drain'] for site in params['allsites']}

        submission = {
                'workflow': workflow,
                'parameters': {
                    'Action': 'acdc',
                    'Reasons': ['AIEH', action.get('description')],
                    'ACDCs': [],
                    'Parameters': {
                        step: {
                            'memory': '',
                            'cores': '',
                            'sites': [site for site in sitelist if site_map.get(site) == 'enabled']
                            }
                        for step, sitelist in params['sitestorun'].items()
                        }
                    }
                }

        check_parameters = submission['parameters']['Parameters']

        if not check_parameters:
            return {}

        for step, step_dict in check_parameters.items():
            if not step_dict['sites']:
                return {}
            if action['tasks'][step]['xrootd'] == 'yes':
                step_dict['xrootd'] = 'enabled'
                step_dict['secondary'] = 'enabled'

        return submission


    @cherrypy.expose
    @cherrypy.tools.json_out()
    def assistance_manual(self):
        return errorutils.assistance_manual()


    @cherrypy.expose
    def check_auto_acdc(self, key):

        if key == serverconfig.config_dict()['actions']['key']:

            aieh_config = serverconfig.config_dict()['aieh']
            user = aieh_config['user']
            dry = aieh_config.get('dry')

            for workflow in errorutils.assistance_manual():

                if self.wkfparams(workflow).get('RequestType', 'Resubmission') == 'Resubmission' \
                        or self.get_status(workflow) != 'none':
                    continue

                params = self.auto_acdc(workflow)

                if params:
                    if dry:
                        print(submission)
                    else:
                        manageactions.submit2(user, [submission])

            self.update_statuses()

        return 'Done'



# pylint: disable=too-many-locals

"""Module to manage actions of WorkflowWebTools.

:author: Daniel Abercrombie <dabercro@mit.edu>
"""

from __future__ import print_function

import time
import datetime
import ssl

import cherrypy
import pymongo

from . import serverconfig
from . import reasonsmanip
from .globalerrors import check_session

def extract_reasons_params(action, **kwargs):
    """Extracts the reasons and parameters for an action from kwargs

    :param str action: The action being asked for by the operator
    :param kwargs: Keywords that are submitted via POST to /submitaction
    :returns: reasons for action, and parameters for the action
    :rtype: list, dict
    """

    old_reasons = reasonsmanip.reasons_list()
    reasons = []
    notupdate = []
    params = {}

    tasks_to_do = kwargs.get('dotasks', [])
    if not isinstance(tasks_to_do, list):
        tasks_to_do = [tasks_to_do]

    for key, item in kwargs.items():

        if 'shortreason' in key:
            short_re = item or reasonsmanip.DEFAULT_SHORT
            long_re = kwargs[key.replace('short', 'long')].strip().replace('\n', '<br>') or item

            reasons.append({
                'short': short_re,
                'long': long_re,
            })

        elif 'selectedreason' in key:

            selectedlist = item if isinstance(item, list) else [item]

            for short in selectedlist:
                if short != "none":
                    notupdate.append({
                        'short': short,
                        'long': old_reasons[short],
                    })

        elif 'param_' in key:
            parameter = '_'.join(key.split('_')[2:])

            if action in ['acdc', 'recovery']:
                default = 'AllSteps' if parameter != 'sites' else 'Ban'
                which_task = kwargs.get('task_%s' % key.split('_')[1], default)

                if tasks_to_do and which_task != default \
                        and which_task not in tasks_to_do:
                    continue

                params[which_task] = params.get(which_task, {})

                if parameter == 'sites' and not isinstance(item, list):
                    item = [item]

                params[which_task].update({parameter: item})

            else:
                params[parameter] = item

    reasonsmanip.update_reasons(reasons)

    return reasons + notupdate, params


def submitaction(user, workflows, action, session=None, **kwargs):
    """Writes the action to Unified and notifies the user that this happened

    :param str user: is the user that submitted the action
    :param str workflows: is the original workflow name or workflows
    :param str action: is the suggested action for Unified to take
    :param cherrypy.Session session: the current session
    :param kwargs: can include various reasons and additional datasets
    :returns: a tuple of workflows, reasons, and params for the action
    :rtype: list, str, list of dicts, dict
    """

    cherrypy.log('args: {0}'.format(kwargs))
    dotasks = kwargs.get('dotasks', [])
    if not isinstance(dotasks, list):
        dotasks = [dotasks]

    reasons, params = extract_reasons_params(action, **kwargs)

    cherrypy.log('Parameters: {0}'.format(params))

    coll = get_actions_collection()

    error_info = check_session(session)

    # Let's define some lambdas to use to filter our ACDCs (sanity checks)
    # First a quick shorthand for workflow parameters
    get_params = lambda wkf: error_info.get_workflow(wkf).get_workflow_parameters()

    # We want request type to be resubmission in our ACDC
    is_resub = lambda wkf: \
        get_params(wkf)['RequestType'] == 'Resubmission'

    # We want our ACDC to be submitted after the original request
    is_new = lambda wkf, workflow: \
        datetime.datetime(*(get_params(wkf)['RequestDate'])) > \
        datetime.datetime(*(get_params(workflow)['RequestDate']))


    if not isinstance(workflows, list):
        workflows = [workflows]

    for workflow in workflows:
        wf_params = dict(params)
        step_list = error_info.get_step_list(workflow)
        short_step_list = ['/'.join(step.split('/')[2:]) for step in step_list]
        # For recovery, get the proper sites and parameters out for each step
        if action in ['acdc', 'recovery']:
            all_steps = wf_params.pop('AllSteps', {})
            banned_sites = wf_params.pop('Ban', {'sites': []})['sites']

            # Fill empty parameters for each step from AllSteps
            for short_step_name, step_name in zip(short_step_list, step_list):
                # Get any existing thing (most likely not there)
                step_params = wf_params.get(short_step_name, {})

                for key, val in all_steps.items():
                    # This also includes if the key value is set but blank
                    if not step_params.get(key):
                        step_params[key] = val

                wf_params[short_step_name] = step_params

                # Set the sites
                if kwargs.get('method', 'Manual') != 'Manual':
                    # Banned sites would show up under 'AllSteps'
                    wf_params[short_step_name]['sites'] = \
                        [site for site in \
                             check_session(session).get_workflow(workflow).\
                             site_to_run(step_name) if site not in banned_sites]

            # Only keep the workflow parameters with steps that occur in given workflow
            wf_params = {key: wf_params[key] for key in wf_params \
                             if key in short_step_list and \
                             (key in dotasks or not dotasks)}

        document = {
            'Action': action,
            'Parameters': wf_params,
            'Reasons': [reason['long'] for reason in reasons],
            'user': user,
            'ACDCs': [wkf for wkf in error_info.get_prepid(
                error_info.get_workflow(workflow).get_prep_id()).get_workflows() \
                          if wkf != workflow and is_resub(wkf) and is_new(wkf, workflow)]
            }

        cherrypy.log('About to insert workflow: %s action: %s' % (workflow, document))

        coll.update_one({'workflow': workflow},
                        {'$set':
                             {'timestamp': int(time.time()),
                              'parameters': document,
                              'acted': 0}},
                        upsert=True)

    return workflows, reasons, params


def submit2(user, documents): # pylint: disable=missing-docstring
    coll = get_actions_collection()

    for document in documents:
        workflow = document['workflow']
        params = document['parameters']
        params['user'] = user

        cherrypy.log('About to insert workflow: %s action: %s' % (workflow, params))

        coll.update_one({'workflow': workflow},
                        {'$set':
                             {'timestamp': int(time.time()),
                              'parameters': params,
                              'acted': 0}},
                        upsert=True)


def get_actions(num_days=None, num_hours=24, acted=0):
    """Get the recent actions to be acted on in dictionary form

    :param int num_days: is the number of days to check for actions
    :param int num_hours: can be used instead of num_days for finer granularity.
                          If ``num_days`` is given, ``num_hours`` is ignored.
    :param int acted: Flag (0, 1, or None) for whether to use acted or not
    :returns: A dictionary of actions, to be rendered as JSON
    :rtype: dict
    """

    if num_days is not None:
        num_hours = num_days * 24

    output = {}
    coll = get_actions_collection()

    age_to_compare = int(time.time()) - num_hours * 3600 if num_hours > 0 else 0

    query = {'timestamp': {'$gt': age_to_compare}, 'acted': acted}
    if acted is None:
        query.pop('acted')

    for match in coll.find(query):
        output[match['workflow']] = match['parameters']

    return output


def get_datetime_submitted(workflow):
    """Get the datetime for a submitted workflow

    :param str workflow: A workflow to get the time submitted
    :returns: A datetime object, or None if the workflow has never been submitted
    :rtype: datetime.datetime or None
    """

    coll = get_actions_collection()

    info = coll.find_one({'workflow': workflow})
    if info:
        return datetime.datetime.fromtimestamp(info['timestamp'])

    return None


def get_acted_workflows(num_days):
    """Get all of the workflows that have actions assigned

    :param int num_days: is the number of past days to check for actions.
                         This speeds up the check while not losing the
                         ability to look farther back in time.
    :returns: a list of workflows acted on
    :rtype: list
    """

    return list(get_actions(num_days, acted=None))


def report_actions(workflows, output=None):
    """Mark actions as acted on

    :param list workflows: is the list of workflows to no longer show
    :param dict output: If set, a reference to a dictionary to update
                        with details of what was and wasn't set to acted
    """
    coll = get_actions_collection()

    if output is not None:
        records = list(coll.find({'workflow': {'$in': workflows}}))
        output['success'] = [record['workflow'] for record in records
                             if record['acted'] == 0]
        output['already_reported'] = [record['workflow'] for record in records
                                      if record['acted'] != 0]

        record_names = [record['workflow'] for record in records]
        output['does_not_exist'] = [wrkf for wrkf in workflows if
                                    wrkf not in record_names]

    coll.update_many({'workflow': {'$in': workflows}, 'acted': 0},
                     {'$set': {'acted': 1}})


def get_actions_collection():
    """Gets the actions collection from MongoDB.

    :returns: the actions collection
    :rtype: pymongo.collection.Collection
    """

    config_dict = serverconfig.config_dict()['actions']
    uri = config_dict.get('uri')
    if uri:
        client = pymongo.MongoClient(uri, ssl_cert_reqs=ssl.CERT_NONE)
    else:
        client = pymongo.MongoClient()

    coll = client[config_dict['database']].actions

    if coll.count() == 0 or 'workflow' not in list(coll.index_information()):
        coll.create_index([('workflow', pymongo.TEXT)],
                          name='workflow', unique=True)

    return coll


def fix_sites(**kwargs):
    """Fix the site lists for tasks that had zero sites.

    :param kwargs: Keywords to be parsed by :py:func:`extract_reasons_params`
    """

    _, params = extract_reasons_params('acdc', **kwargs)

    coll = get_actions_collection()

    for task, value in params.items():
        split_task = task.split('/')
        workflow = split_task[1]
        subtask = '/'.join(split_task[2:])

        output = coll.find_one({'workflow': workflow})['parameters']
        output['Parameters'][subtask]['sites'] = value['sites']

        coll.update_one({'workflow': workflow},
                        {'$set': {'parameters': output}})

    print(params)



#! /usr/bin/env python

import unittest
import json
import shutil
import os
import sys

import cmstoolbox.webtools
cmstoolbox.webtools.get_json = lambda *a, **k: {}

import workflowwebtools.serverconfig as sc
sc.LOCATION = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    'config.yml')

import cmstoolbox._loadtestpath

import update_history as uh
import workflowwebtools.reasonsmanip as rm
import workflowwebtools.manageactions as ma
import workflowwebtools.globalerrors as ge

from workflowwebtools.paramsregression import convert_to_dense

from workflowwebtools.workflowinfo import WorkflowInfo

class TestGlobalError(unittest.TestCase):

    testdat = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        'testdat.json')

    dictionary = {
        '/test1/a/1': {
            'errors': {
                'row1': {'col1' : 1}, 
                'row2': {'col2': 1, 
                         'col3': 1}
                },
            'sub': {},
            'total': 3
            },
        '/test1/a/2': {
            'errors': {
                'row1': {'col1' : 1,
                         'col2' : 1}, 
                'row2': {'col3': 1}
                },
            'sub': {},
            'total': 3
            },
        '/test2/a/1': {
            'errors': {
                'row2': {'col1': 1}
                },
            'sub': {},
            'total': 1
            }
        }

    def test_grouping(self):
        group_by = lambda subtask: subtask.split('/')[1]
        check_this = ge.group_errors(self.dictionary, group_by)

        self.assertEqual(len(check_this.keys()), 2)
        self.assertTrue('test1' in check_this.keys())
        self.assertTrue('test2' in check_this.keys())
        self.assertEqual(check_this['test1']['errors'], {'row1': {'col1': 2, 'col2': 1}, 'row2': {'col2': 1, 'col3': 2}})
        self.assertEqual(check_this['test2']['errors'], {'row2': {'col1': 1}})
        self.assertEqual(check_this['test1']['sub']['/test1/a/1'], self.dictionary['/test1/a/1'])

    def test_steplist(self):
        info = ge.ErrorInfo(self.testdat)

        self.assertEqual(info.get_step_list('test1'), ['/test1/a/1', '/test1/a/2'])
        self.assertEqual(info.get_step_list('test2'), ['/test2/a/1'])
        self.assertFalse(info.get_step_list('test3'))

    def test_reset(self):
        info = ge.ErrorInfo(self.testdat)
        # Let's load the new one
        info.data_location = self.testdat.replace('.json', '2.json')
        self.assertFalse(info.get_step_list('test3'))
        info.teardown()
        info.setup()
        self.assertEqual(info.get_step_list('test3'), ['/test3/test/2'])
        self.assertFalse(info.get_step_list('test1'))


class TestClusteringAndReasons(unittest.TestCase):

    errors = {
        '/test1/a/1': {
            '1': {
                'site_a': 100
                },
            '2': {
                'site_a': 10,
                'site_b': 100
                },
            },
        '/test2/a/2': {
            '1': {
                'site_a': 90
                },
            '2': {
                'site_a': 10,
                'site_b': 100
                },
            },
        '/test3/b/a': {
            '3': {
                'site_c': 50
                }
            },
        '/crash/time/haha': {
            'NotANumber': {
                'site_c': 0
                }
            }
        }

    should_cluster = {
        'test1' : ['test2'],
        'test2' : ['test1'],
        'test3' : []
        }

    def setUp(self):
        out_name = sc.all_errors_path()

        if not os.path.exists(os.path.dirname(out_name)):
            os.makedirs(os.path.dirname(out_name))

        with open(sc.all_errors_path(), 'w') as output:
            json.dump(self.errors, output)

        if sc.workflow_history_path() != sc.all_errors_path():
            uh.main(sc.all_errors_path())

        ge.check_session(None).data_location = out_name
        ge.check_session(None).setup()

        for wkf in ge.check_session(None).return_workflows():
            file_name = WorkflowInfo(wkf).cache_filename('workflow_params')
            dirname = os.path.dirname(file_name)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            with open(file_name, 'w') as cache:
                json.dump({}, cache)

    def tearDown(self):
        os.remove(sc.workflow_history_path())
        if sc.workflow_history_path() != sc.all_errors_path():
            os.remove(sc.all_errors_path())

        for wkf in ge.check_session(None).return_workflows():
            file_name = WorkflowInfo(wkf).cache_filename('workflow_params')
            os.remove(file_name)

        ge.check_session(None).teardown()

    def test_updatehistory(self):
        import workflowwebtools.globalerrors as ge

        self.assertEqual(ge.ErrorInfo(sc.all_errors_path()).info[1:],
                         ge.ErrorInfo(sc.workflow_history_path()).info[1:],
                         'Update workflow script did not create equivalent database')

    def test_clusterer(self):
        import workflowwebtools.globalerrors as ge
        import workflowwebtools.clusterworkflows as cw

        clusterer = cw.get_clusterer(sc.workflow_history_path())

        for workflow in ge.GLOBAL_INFO.return_workflows():
            self.assertEqual(cw.get_clustered_group(workflow, clusterer),
                             self.should_cluster[workflow],
                             'Clustering acting unexpectedly.')

    def test_steptable(self):
        # This isn't particularly well written,
        # but we should expect a table with
        # rows corresponding to error codes 1, 2, 3 and
        # columns corresponding to site names site_a, site_b, site_c

        allmap = ge.check_session(None).get_allmap()

        for step in self.errors:
            error_table = ge.get_step_table(step)

            # One by hand because weird stuff is happening while writing this test
            if step == '/test1/a/1':
                self.assertEqual(error_table[0][0], 100)

            self.assertTrue(error_table)
            for row in error_table:
                self.assertTrue(row)

            # Check number of rows and errors
            self.assertEqual(len(allmap['errorcode']), len(error_table))
            for e_index, error in enumerate(allmap['errorcode']):
                # Make sure all of the sites have a column
                self.assertEqual(len(allmap['sitename']), len(error_table[e_index]))
                for s_index, site in enumerate(allmap['sitename']):
                    self.assertEqual(error_table[e_index][s_index],
                                     self.errors[step].get(str(error), {}).get(site, 0))

    def test_sparsetodense(self):
        dense = {}
        sparse = {}
        for step in self.errors:
            sparse[step] = ge.get_step_table(step, sparse=True)
            dense[step] = ge.get_step_table(step)

        to_dense = convert_to_dense(sparse, self.errors.keys())

        for step in self.errors:
            self.assertEqual(dense[step], to_dense[step])


class TestReasons(unittest.TestCase):

    reasons = [
        {
            'short': 'short reason 1',
            'long': 'secretly long reason'
        },
        {
            'short': 'short reason 2',
            'long': 'another secretly long reason'
        }
    ]

    def setUp(self):
        if os.path.exists('reasons.db'):
            os.remove('reasons.db')

        print('before', rm.reasons_list())

        rm.update_reasons(self.reasons)

        print('after', rm.reasons_list())


    def tearDown(self):
        os.remove('reasons.db')

    def test_reasons(self):

        self.assertRaises(TypeError, rm.update_reasons, 'test')
        self.assertRaises(KeyError, rm.update_reasons, [{'wrong': 'key'}])
        self.assertEqual(rm.reasons_list(),
                         {reas['short']: reas['long'] for reas in self.reasons},
                         'Reasons list return is not same as sent,\n\n%s\n\n%s' %
                         (rm.reasons_list(), {reas['short']: reas['long'] for reas in self.reasons}))


class TestActions(unittest.TestCase):

    reasons1 = [
        {
            'short': 'short reason 1',
            'long': 'long reason 1'
        }
    ]

    reasons2 = {
        'shortreason0': 'short reason 2',
        'longreason0': 'long reason 2',
        'shortreason4': 'short reason 3',
        'longreason4': 'long reason 3',
        'selectedreason': 'short reason 1',
    }

    request_base = {
        'workflows': 'test_workflow',
        'task_0': 'task/0',
        'task_1': 'task/1',
    }

    def extend_request(self, request):
        output = request
        output.update(self.reasons2)
        output.update(self.request_base)

        return output

    def setUp(self):
        rm.update_reasons(self.reasons1)

        if ma.get_actions_collection().count() != 0:
            print('Test database not empty, abort!!')
            exit(123)

        file_name = WorkflowInfo(self.request_base['workflows']).cache_filename('workflow_params')
        dirname = os.path.dirname(file_name)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        with open(file_name, 'w') as cache:
            json.dump({}, cache)

    def tearDown(self):
        os.remove('reasons.db')
        ma.get_actions_collection().drop()

        file_name = WorkflowInfo(self.request_base['workflows']).cache_filename('workflow_params')
        os.remove(file_name)

    def run_test(self, request, params_out):

        wf, reasons, params = ma.submitaction('test', **request)

        self.assertEqual(wf, [self.request_base['workflows']],
                         'Workflows is not expected output')
        
        self.assertEqual(reasons,
                         [{'short': 'short reason %i' % i, 'long': 'long reason %i' % i} for i in [2, 3, 1]],
                         'Output reasons are not what are expected')

        self.assertEqual(params, params_out,
                         'Parameters out are not expected')

    def test_clone(self):
        request = self.extend_request({
                'action': 'clone',
                'param_0_test': 'test_param'
                })
        self.run_test(request, {'test': 'test_param'})

    def test_acdc(self):
        request = self.extend_request({
                'action': 'acdc',
                'param_0_test': 'test_param_0',
                'param_1_test': 'test_param_1',
                })
        self.run_test(request, {'task/0': {'test': 'test_param_0'},
                                'task/1': {'test': 'test_param_1'}})

    def test_investigate(self):
        request = self.extend_request({
                'action': 'investigate',
                'param_0_test': 'test_param'
                })
        self.run_test(request, {'test': 'test_param'})

class TestBlankReasons(TestActions):
    def test_blank_short(self):
        request = {
            'workflows': 'test_workflow',
            'action': 'clone',
            'shortreason0': '',
            'longreason0': 'Long reason'
            }

        wf, reasons, params = ma.submitaction('test', **request)
        self.assertEqual(reasons[0]['short'], rm.DEFAULT_SHORT)

    def test_blank_long(self):
        request = {
            'workflows': 'test_workflow',
            'action': 'clone',
            'shortreason0': 'Short Reason',
            'longreason0': ''
            }

        wf, reasons, params = ma.submitaction('test', **request)
        self.assertEqual(reasons[0]['short'], reasons[0]['long'])

if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python

"""
Test the different responses that reporting an action could do
"""

import os
import time
import unittest

import cmstoolbox.webtools
cmstoolbox.webtools.get_json = lambda *a, **k: {}

from workflowwebtools import serverconfig
serverconfig.LOCATION = os.path.join(
    os.path.dirname(os.path.realpath(__file__)),
    'config.yml')

from workflowwebtools import manageactions

class TestReportResponse(unittest.TestCase):
    new_workflows = ['test_workflow_1', 'test_workflow_2', 'new_test']
    old_workflows = ['old_workflow_1', 'old_workflow_2']
    fake_workflows = ['does_not_exist', 'not_workflow']

    def setUp(self):
        self.output = {}
        self.coll = manageactions.get_actions_collection()
        # Check that test database is empty
        if self.coll.count() != 0:
            print('Test database not empty, abort!!')
            exit(123)

        now = time.time()
        # Add the workflows that we want to say exists
        for acted, workflows in [(0, self.new_workflows), (1, self.old_workflows)]:
            for wf in workflows:
                self.coll.insert_one({'workflow': wf, 'acted': acted, 'timestamp': now, 'parameters': {}})

    def tearDown(self):
        self.coll.drop()

    def test_setup(self):
        self.assertEqual(self.coll.count(), len(self.new_workflows + self.old_workflows))
        self.assertEqual(sorted(self.new_workflows),
                         sorted(manageactions.get_actions(acted=0).keys()))
        self.assertEqual(sorted(self.old_workflows),
                         sorted(manageactions.get_actions(acted=1).keys()))

    def test_new(self):
        manageactions.report_actions(self.new_workflows, self.output)

        self.assertEqual(sorted(self.new_workflows), sorted(self.output['success']))
        self.assertFalse(self.output['already_reported'])
        self.assertFalse(self.output['does_not_exist'])

    def test_old(self):
        manageactions.report_actions(self.old_workflows, self.output)

        self.assertFalse(self.output['success'])
        self.assertEqual(sorted(self.old_workflows), sorted(self.output['already_reported']))
        self.assertFalse(self.output['does_not_exist'])

    def test_not_exist(self):
        manageactions.report_actions(self.fake_workflows, self.output)

        self.assertFalse(self.output['success'])
        self.assertFalse(self.output['already_reported'])
        self.assertEqual(sorted(self.fake_workflows), sorted(self.output['does_not_exist']))

    def test_everything(self):
        manageactions.report_actions(self.new_workflows + 
                                     self.old_workflows +
                                     self.fake_workflows, self.output)

        self.assertEqual(sorted(self.new_workflows), sorted(self.output['success']))
        self.assertEqual(sorted(self.new_workflows), sorted(self.output['success']))
        self.assertEqual(sorted(self.fake_workflows), sorted(self.output['does_not_exist']))


if __name__ == '__main__':
    unittest.main()



#pylint: skip-file

"""
Module containing and returning information about workflows.

:authors: Daniel Abercrombie <dabercro@mit.edu>
"""

from __future__ import print_function

import os
import re
import json
import time
import datetime
import threading

from collections import defaultdict
from functools import wraps

from cmstoolbox.webtools import get_json
from cmstoolbox.sitereadiness import site_list

from . import serverconfig

def cached_json(attribute, timeout=None):
    """
    A decorator for caching dictionaries in local files.

    :param str attribute: The key of the :py:class:`WorkflowInfo` cache to
                          set using the decorated function.
    :param int timeout: The amount of time before refreshing the JSON file, in seconds.
    :returns: Function decorator
    :rtype: func
    """

    def cache_decorator(func):
        """
        The actual decorator (since decorator takes an argument)

        :param func func: A function to modify
        :returns: Decorated function
        :rtype: func
        """

        @wraps(func)
        def function_wrapper(self, *args, **kwargs):
            """
            Executes the :py:class:`WorkflowInfo` function

            :returns: Output of the originally decorated function
            :rtype: dict
            """
            tmout = timeout or serverconfig.config_dict()['cache_refresh'].get(attribute)

            if not os.path.exists(self.cache_dir):
                os.mkdir(self.cache_dir)

            self.cachelock.acquire()
            if attribute not in self.cachelocks:
                self.cachelocks[attribute] = threading.Lock()

            self.cachelocks[attribute].acquire()
            self.cachelock.release()

            check_var = self.cache.get(attribute)

            if check_var is None:
                file_name = self.cache_filename(attribute)
                if os.path.exists(file_name) and \
                        (tmout is None or time.time() - tmout < os.stat(file_name).st_mtime):
                    try:
                        with open(file_name, 'r') as cache_file:
                            check_var = json.load(cache_file)
                    except ValueError:
                        print('JSON file no good. Deleting %s. Try again later.' % file_name)
                        os.remove(file_name)

                # If still None, call the wrapped function
                if check_var is None:
                    check_var = func(self, *args, **kwargs)
                    with open(file_name, 'w') as cache_file:
                        json.dump(check_var, cache_file)

                self.cache[attribute] = check_var

            self.cachelocks[attribute].release()

            return check_var or {}

        return function_wrapper

    return cache_decorator


def list_workflows(status):
    """
    Get the list of workflows currently in a given status.
    For a list of valid requests, visit the
    `Request Manager Interface <https://cmsweb.cern.ch/reqmgr2/>`_.

    :param str status: The status of the workflow lists being looked for
    :returns: A list of workflows matching the status
    :rtype: list
    """

    request = get_json('cmsweb.cern.ch', '/reqmgr2/data/request',
                       params={'status': status, 'detail': 'false'},
                       use_cert=True)

    return request['result']


def errors_for_workflow(workflow, url='cmsweb.cern.ch'):
    """
    Get the useful status information from a workflow

    :param str workflow: the name of the workflow request
    :param str url: the base url to find the information at
    :returns: a dictionary containing error codes in the following format::

              {step: {errorcode: {site: number_errors}}}

    :rtype: dict
    """

    result = get_json(url,
                      '/wmstatsserver/data/jobdetail/%s' % workflow,
                      use_cert=True)

    output = {}

    if not result['result']:
        return output

    for step, stepdata in result['result'][0].get(workflow, {}).items():
        errors = {}
        for code, codedata in stepdata.get('jobfailed', {}).items():
            sites = {}
            for site, sitedata in codedata.items():
                if sitedata['errorCount']:
                    sites[site] = sitedata['errorCount']

            if sites:
                errors[code] = sites

        if errors:
            output[step] = errors

    return output

def explain_errors(workflow, errorcode):
    """
    Get example errors for a given workflow and errorcode

    :param str workflow: is the workflow name
    :param str errorcode: is the error code
    :returns: a dict of log snippets from different sites.
    :rtype: list
    """

    result = get_json('cmsweb.cern.ch',
                      '/wmstatsserver/data/jobdetail/%s' % workflow,
                      use_cert=True)

    output = []

    if not result['results']:
        return output

    for stepdata in result['result'][0].get(workflow, {}).values():
        for sitedata in stepdata.get('jobfailed', {}).get(errorcode, {}).values():
            for samples in sitedata['samples'][0]['errors'].values():

                output.extend(samples)

    return output


class Info(object):
    """
    Implements shared operations on the cache
    """

    def __init__(self):
        # Stores things using the cached_json decorator
        self.cache = {}
        self.cache_dir = os.path.join(os.environ.get('TMPDIR', '/tmp'), 'workflowinfo')
        self.bak_dir = os.path.join(self.cache_dir, 'bak')
        self.cachelock = threading.Lock()
        self.cachelocks = {}

    def __str__(self):
        pass

    def cache_filename(self, attribute):
        """
        Return the name of the file for caching

        :param str attribute: The information to store in the file
        :returns: The full file name to store the cache
        :rtype: str
        """
        return os.path.join(self.cache_dir, '%s_%s.cache.json' % (self, attribute))

    def reset(self):
        """
        Reset the cache for this object and clear out the files.
        """
        print('Reseting %s' % self)

        if not os.path.exists(self.bak_dir):
            os.mkdir(self.bak_dir)

        for attribute in self.cache:
            cache_file = self.cache_filename(attribute)
            if os.path.exists(cache_file):
                os.rename(cache_file, cache_file.replace(self.cache_dir, self.bak_dir))

        self.cache.clear()


class WorkflowInfo(Info):
    """
    Class that holds methods for accessing various information about a workflow.
    """

    def __init__(self, workflow, url='cmsweb.cern.ch'):
        """
        Initialize the workflow info class

        :param str workflow: is the name of the workflow
        :param str url: is the url to fetch information from
        """

        super(WorkflowInfo, self).__init__()
        self.workflow = workflow
        self.url = url

        # Is set first time get_explanation() is called
        self.explanations = None

    def __str__(self):
        return 'workflowinfo_%s' % self.workflow

    @cached_json('workflow_params')
    def get_workflow_parameters(self):
        """
        Get the workflow parameters from ReqMgr2, or returns a cached value.
        See the `ReqMgr 2 wiki <https://github.com/dmwm/WMCore/wiki/reqmgr2-apis>`_
        for more details.

        :returns: Parameters for the workflow from ReqMgr2.
        :rtype: dict
        """

        try:
            result = get_json(self.url,
                              '/reqmgr2/data/request',
                              params={'name': self.workflow},
                              use_https=True, use_cert=True)

            for params in result['result']:
                for key, item in params.items():
                    if key == self.workflow:
                        return item

        except Exception as error:
            print('Failed to get from reqmgr', self.workflow)
            print(str(error))

        return None


    @cached_json('errors')
    def get_errors(self, get_unreported=False):
        """
        A wrapper for :py:func:`errors_for_workflow` if you happen to have
        a :py:class:`WorkflowInfo` object already.

        :param bool get_unreported: Get the unreported errors from ACDC server
        :returns: a dictionary containing error codes in the following format::

                  {step: {errorcode: {site: number_errors}}}

        :rtype: dict
        """

        output = errors_for_workflow(self.workflow, self.url)

        if get_unreported:
            acdc_server_response = get_json(
                'cmsweb.cern.ch', '/couchdb/acdcserver/_design/ACDC/_view/byCollectionName',
                {'key': '"%s"' % self.workflow, 'include_docs': 'true', 'reduce': 'false'},
                use_cert=True)

            for row in acdc_server_response['rows']:
                task = row['doc']['fileset_name']

                new_output = output.get(task, {})
                new_errorcode = new_output.get('NotReported', {})
                for file_replica in row['doc']['files'].values():
                    for site in file_replica['locations']:
                        new_errorcode[site] = 0

                new_output['NotReported'] = new_errorcode
                output[task] = new_output

        for step in list(output):
            if True in [(steptype in step) for steptype in ['LogCollect', 'Cleanup']]:
                output.pop(step)

        return output

    @cached_json('reqdetail')
    def _get_reqdetail(self):
        """
        Get the request detail from the wmstatsserver

        :returns: The request detail json from the server or cache
        :rtype: dict
        """

        reqDetail = {self.workflow : {}}
        raw =  get_json(self.url,
                        '/wmstatsserver/data/request/%s' % self.workflow,
                        use_cert=True)

        result = raw.get('result', [])
        if not result: return reqDetail

        reqDetail[self.workflow] = result[0].get(self.workflow, {})

        return reqDetail

    def get_failure_rate(self):
        """
        Get failure rate if you happen to have a `WorkflowInfo` object

        :returns: workflow's failure rate

        :rtype: float
        """

        result = self._get_reqdetail()

        frate = 0.
        wf_agents = result.get(self.workflow, {}).get('AgentJobInfo', {})
        if not wf_agents:
            return frate

        nsuccess = 0
        nfailure = 0
        for agent, agentdata in wf_agents.items():
            status = agentdata.get('status', {})
            if not status: continue

            nsuccess += status.get('success', 0)

            for ftype, num in status.get('failure', {}).items():
                nfailure += num

        try:
            frate = float(nfailure)/(nfailure+nsuccess)
        except ZeroDivisionError:
            pass

        return frate

    def sum_errors(self):
        """
        :returns: The total number of errors reported by this workflow
        :rtype: int
        """

        errors = self.get_errors(True)
        return sum([num for codes in errors.values() for sites in codes.values()
                    for num in sites.values()])


    @cached_json('recovery_info')
    def get_recovery_info(self):
        """
        Get the recovery info for this workflow.

        :returns: a dictionary containing the information used in recovery.
                  The keys in this dictionary are arranged like the following::

                  { task: { 'sites_to_run': list(sites), 'missing_to_run': int() } }

        :rtype: dict
        """

        recovery_info = {}

        docs = get_json(self.url,
                        '/couchdb/acdcserver/_design/ACDC/_view/byCollectionName',
                        params={'key': '"%s"' % self.workflow,
                                'include_docs': 'true',
                                'reduce': 'false'},
                        use_cert=True)

        recovery_docs = [row['doc'] for row in docs.get('rows', [])]
        site_white_list = set(self.get_workflow_parameters()['SiteWhitelist'])

        for doc in recovery_docs:
            task = doc['fileset_name']
            # For each task, we have the following keys:
            # sites - a set of sites that the recovery docs say to run on.
            for replica, info in doc['files'].items():
                # For fake files, just return the site whitelist
                if replica.startswith('MCFakeFile'):
                    locations = site_white_list
                else:
                    locations = set(info['locations'])

                vals = recovery_info.get(task, {})
                if not vals:
                    recovery_info[task] = {}

                recovery_info[task]['sites_to_run'] = \
                    list(set(vals.get('sites_to_run', set())) | locations)
                recovery_info[task]['missing_to_run'] = \
                    (vals.get('missing_to_run', 0) + info['events'])

        return recovery_info


    def site_to_run(self, task):
        """
        Gets a list of sites that a task in the workflow can run at

        :param str task: The full name of the task to find sites for
        :returns: a list of site to run at
        :rtype: list
        """

        site_set = self.get_recovery_info().get(task, {}).get('sites_to_run', [])
        out_list = []
        all_site_list = site_list()

        for site in site_set:
            if site.startswith('T0_') or site.endswith('_MSS') or site.endswith('_Export'):
                continue

            clean_site = re.sub(r'_(ECHO_)?(Disk)$', '', site)
            if clean_site not in out_list and clean_site and \
                    clean_site in all_site_list:
                out_list.append(clean_site)

        out_list.sort()

        return out_list

    @cached_json('jobdetail')
    def _get_jobdetail(self):
        """
        Get the jobdetail from the wmstatsserver

        :returns: The job detail json from the server or cache
        :rtype: dict
        """

        return get_json(self.url,
                        '/wmstatsserver/data/jobdetail/%s' % self.workflow,
                        use_cert=True)

    def get_explanation(self, errorcode, step=''):
        """
        Gets a list of error logs for a given error code.

        :param str errorcode: The error code to explain
        :param str step: The full name of the step to return explanations from
        :returns: list of error logs
        :rtype: list
        """

        if self.explanations is None:
            self.explanations = defaultdict(lambda: defaultdict(lambda: []))
            result = self._get_jobdetail()
            for stepname, stepdata in result['result'][0].get(self.workflow, {}).items():
                # Get the errors from both 'jobfailed' and 'submitfailed' details
                for error, site in [(error, site) for status in ['jobfailed', 'submitfailed'] \
                                        for error, site in stepdata.get(status, {}).items()]:
                    if error == '0':
                        continue

                    for sitename, samples in site.items():
                        for detail in [values for sample in samples['samples']
                                       for errs in sample['errors'].values()
                                       for values in errs]:

                            self.explanations[error][stepname].append('\n\n'.join(
                                ['Site name: %s' % sitename,
                                 '%s (Exit code: %s)' % (detail['type'], detail['exitCode']),
                                 detail['details']]))

        explain = self.explanations.get(errorcode, {'': ['No info for this error code']})

        if step in explain.keys():
            return explain[step]

        return [val for expl in explain.values() for val in expl]

    def get_prep_id(self):
        """
        :returns: the PrepID for this workflow
        :rtype: str
        """

        return str(self.get_workflow_parameters().get('PrepID', 'NoPrepID'))

    def get_monitoring_info(self):
        """
        :returns: the information to send to CMSMONIT
        :rtype: dict
        """
        # Dummy call to get self.explanations filled
        self.get_explanation(0)

        return {
            'errors': self.get_errors(True),
            'prepID': self.get_prep_id(),
            'params': self.get_workflow_parameters(),
            'recovery': self.get_recovery_info(),
            'logs': self.explanations
            }


class PrepIDInfo(Info):
    """
    A class that just holds a small amount of information about a given PrepID.
    """

    def __init__(self, prep_id, url='cmsweb.cern.ch'):
        super(PrepIDInfo, self).__init__()
        self.prep_id = prep_id
        self.url = url

    def __str__(self):
        return 'prepIDinfo_%s' % self.prep_id

    @cached_json('requests')
    def get_requests(self):
        """
        :returns: The requests for the Prep ID from ReqMgr2 API
        :rtype: dict
        """
        if self.prep_id == 'NoPrepID':
            return None

        result = get_json(self.url, '/reqmgr2/data/request',
                          params={'prep_id': self.prep_id, 'detail': 'true'},
                          use_cert=True)

        if not result['result']:
            return None

        return result['result'][0]

    def get_workflows_requesttime(self):
        """
        :returns: A list of tuples containing (workflow name, timestamp of request)
        :rtype: list
        """
        request = self.get_requests()

        return [(workflow, time.mktime(datetime.datetime(*value['RequestDate']).timetuple())) \
                    for workflow, value in request.items()]

    def get_workflows(self):
        """
        :returns: A list of tuples containing (workflow name, timestamp of request)
        :rtype: list
        """

        return [
            workflow for workflow, _ in
            self.get_workflows_requesttime()
        ]



"""
This module provides lists of errors per pie variable

:author: Daniel Abercrombie
"""


from .globalerrors import list_matching_pievars
from .globalerrors import check_session


def listworkflows(error_code, site_name, workflow, session=None):
    """
    Gives back a list of tuples containing pie variables and the number of errors
    that matches a given error_code and site_name.

    :param int error_code: The error code that we want errors for
    :param str site_name: The site name that we want errors for
    :param str workflow: The workflow that we want errors for
    :param cherrypy.Session session: holds the session information
    :returns: List of tuples of pie variable and errors, sorted by number of errors
    :rtype: list
    """

    output_dict = {}

    if not workflow:
        for step, numerrors in list_matching_pievars('stepname', error_code, site_name, session):
            wkf = step.split('/')[1]
            output_dict[wkf] = output_dict.get(wkf, 0) + numerrors

    else:
        # Click on step piechart
        if len(workflow.split('/')) > 1:

            if not error_code:
                pievar = 'errorcode'
                col = site_name
            elif not site_name:
                pievar = 'sitename'
                col = error_code
            else:
                return []

            for key, numerrors in list_matching_pievars(pievar, workflow, col, session):
                output_dict[key] = output_dict.get(key, 0) + numerrors

        else:
            # Click on workflow (not step)
            info = check_session(session)
            if workflow in info.return_workflows():
                nextlist = info.get_step_list(workflow)
            # Otherwise, is hopefully a PrepID
            else:
                nextlist = info.get_prepid(workflow).get_workflows()

            for step in nextlist:
                for key, numerrors in listworkflows(error_code, site_name, step, session):
                    output_dict[key] = output_dict.get(key, 0) + numerrors


    return sorted(output_dict.items(), key=lambda x: x[1], reverse=True)



"""Simple utils functions that do not rely on a session.

These are separated from globalerrors, since we do not need
to generate an ErrorInfo instance in other applications.

:author: Daniel Abercrombie <dabercro@mit.edu>
"""


import os
import json
import re
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse # pylint: disable=import-error

import validators
import cherrypy
import cx_Oracle

from cmstoolbox import sitereadiness
from cmstoolbox.webtools import get_json

from . import workflowinfo
from . import serverconfig

def errors_from_list(workflows):
    """
    :param list workflows: A list of workflows that are in assistance-manual
    :returns: The errors for the workflows
    :rtype: dict
    """
    indict = {}

    for workflow in workflows:
        base = workflowinfo.WorkflowInfo(workflow)
        prep_id = base.get_prep_id()
        for wkf in set(workflowinfo.PrepIDInfo(prep_id).get_workflows()):
            indict.update(
                workflowinfo.WorkflowInfo(wkf).get_errors(get_unreported=True)
                )

    return indict


def assistance_manual():
    """
    Returns the workflows in 'assistance-manual'

    :returns: list of workflows
    :rtype: list
    """
    config_dict = serverconfig.config_dict()

    if 'oracle' in config_dict:
        try:
            oracle_db_conn = cx_Oracle.connect(*config_dict['oracle'])
            oracle_cursor = oracle_db_conn.cursor()
            oracle_cursor.execute(
                "SELECT NAME FROM CMS_UNIFIED_ADMIN.workflow WHERE lower(STATUS) = 'assistance-manual'") # pylint:disable=line-too-long
            wkfs = [row for row, in oracle_cursor]
            oracle_db_conn.close()
            return wkfs

        except cx_Oracle.DatabaseError:
            pass

    return []


def open_location(data_location):
    """
    This function assumes that the contents of the location is in JSON format.
    It opens the data location and returns the dictionary.

    :param str data_location: The location of the file or url
    :returns: information in the JSON file
    :rtype: dict
    """
    config_dict = serverconfig.config_dict()

    if 'oracle' in config_dict:
        try:
            oracle_db_conn = cx_Oracle.connect(*config_dict['oracle'])
            oracle_cursor = oracle_db_conn.cursor()
            oracle_cursor.execute(
                "SELECT NAME FROM CMS_UNIFIED_ADMIN.workflow WHERE lower(STATUS) LIKE '%manual%'")
            wkfs = [row for row, in oracle_cursor]
            oracle_db_conn.close()
            return errors_from_list(wkfs)

        except cx_Oracle.DatabaseError:
            return None

    raw = None

    if os.path.isfile(data_location):
        with open(data_location, 'r') as input_file:
            raw = json.load(input_file)

    elif validators.url(data_location):
        components = urlparse.urlparse(data_location)

        # Anything we need for the Shibboleth cookie could be in the config file
        cookie_stuff = config_dict['data']

        raw = get_json(components.netloc, components.path,
                       use_https=True,
                       cookie_file=cookie_stuff.get('cookie_file'),
                       cookie_pem=cookie_stuff.get('cookie_pem'),
                       cookie_key=cookie_stuff.get('cookie_key'))

    if raw is None:
        return raw

    if not (raw and isinstance(raw[list(raw)[0]], list)):
        return raw

    return errors_from_list([
        workflow for workflow, statuses in raw.items()
        if True in ['manual' in status for status in statuses]
    ])


def get_list_info(status_list):
    """
    Get the list of workflows that match the statuses listed
    via :py:mod:`workflowinfo`.

    :param list status_list: The list of workflow statuses to get the info for
    :returns: The workflow info dictionary, which matches the format of
              the unified all_errors.json.
    :rtype: dict
    """

    indict = {}
    for workflow in status_list:
        indict.update(
            workflowinfo.WorkflowInfo(workflow).get_errors(get_unreported=True)
            )

    return indict


def add_to_database(curs, data_location):
    """Add data from a file to a central database through the passed cursor

    :param sqlite3.Cursor curs: is the cursor to the database
    :param data_location: If a string, this
         is the location of the file
         or url of data to add to the database.
         This should be in JSON format, and if a local file does not exist,
         a url will be assumed. If the url is invalid,
         an empty database will be returned.
         If a list, it's a list of status to get workflows from wmstats.
    :type data_location: str or list
    """

    indict = get_list_info(data_location) \
        if isinstance(data_location, list) else \
        (open_location(data_location) or {})

    number_added = 0

    for stepname, errorcodes in indict.items():
        if 'LogCollect' in stepname or 'Cleanup' in stepname:
            continue

        for errorcode, sitenames in errorcodes.items():
            if errorcode == 'NotReported':
                errorcode = '-1'

            elif not re.match(r'\d+', errorcode):
                continue

            for sitename, numbererrors in sitenames.items():
                numbererrors = numbererrors or int(errorcode == '-1')

                if numbererrors:
                    full_key = '_'.join([stepname, sitename, errorcode])
                    if not list(curs.execute(
                            'SELECT EXISTS(SELECT 1 FROM workflows WHERE fullkey=? LIMIT 1)',
                            (full_key,)))[0][0]:
                        number_added += 1
                        curs.execute('INSERT INTO workflows VALUES (?,?,?,?,?,?)',
                                     (full_key, stepname, errorcode,
                                      sitename, numbererrors,
                                      sitereadiness.site_readiness(sitename)))

    # This is to prevent the ErrorInfo objects from locking the database
    if 'conn' in dir(curs):
        curs.conn.commit()

    if number_added:
        cherrypy.log('Number of points added to the database: %i' % number_added)


def create_table(curs):
    """Create the workflows error table with the proper format
    :param sqlite3.Cursor curs: is the cursor to the database
    """

    curs.execute(
        'CREATE TABLE workflows (fullkey varchar(1023) UNIQUE, '
        'stepname varchar(255), errorcode int, '
        'sitename varchar(255), numbererrors int, '
        'sitereadiness varchar(15))')
    # Hopefully this makes lookups faster
    curs.execute('CREATE INDEX composite_index ON workflows (stepname, errorcode, sitename)')



#!/usr/bin/env python
from __future__ import print_function

import os
import re
import sys
import json
import time
import gzip
import shutil
import threading
from collections import defaultdict

import yaml
import cx_Oracle
from workflowwebtools import workflowinfo
from workflowwebtools import errorutils


def save_json(json_obj, filename='tmp', gzipped=False):
    """
    save json object to a local formatted text file, for debug

    :param dict json_obj: the json object
    :param str filename: the base name of the file to be saved
    :param bool gzipped: if gzip output document, default is False
    :returns: full filename

    :rtype: str
    """

    fn = "{}.json".format(filename)
    msg = json.dumps(json_obj, sort_keys=True, indent=4, separators=(',', ': '))

    if gzipped:
        fn += '.gz'
        with gzip.open(fn, 'wb') as f:
            f.write(msg.encode())
    else:
        with open(fn, 'w') as f:
            f.write(msg)

    return fn


def get_yamlconfig(configPath):
    '''
    get a dict of config file (YAML) pointed by configPath.

    :param str configPath: path of config file
    :returns: dict of config

    :rtype: dict
    '''

    if not os.path.isfile(configPath):
        return {}

    try:
        return yaml.load(open(configPath).read(), Loader=yaml.FullLoader)
    except:
        return {}


def invalidate_caches(cacheDir=None):
    '''
    remove json caches pointed by cacheDir

    :param str cache_dir: path of caching directory
    :returns: None
    '''

    cache_dir = cacheDir or os.path.join(
        os.environ.get('TMPDIR', '/tmp'), 'workflowinfo')

    try:
        shutil.rmtree(cache_dir)
    except:
        print('Fail to remove caches: ', cache_dir)
        pass

    try:
        os.mkdir(cache_dir)
    except:
        pass


def get_workflowlist_from_db(config, queryCmd):
    '''
    get a list of workflows from oracle db from a config dictionary which has a ``oracle`` key.

    :param dict config: config dictionary
    :param str queryCmd: SQL query command
    :returns: list of workflow names that are LIKE running

    :rtype: list
    '''

    if 'oracle' not in config:
        return []

    oracle_db_conn = cx_Oracle.connect(
        *config['oracle'])  # pylint:disable=c-extension-no-member
    oracle_cursor = oracle_db_conn.cursor()
    oracle_cursor.execute(queryCmd)
    wkfs = [row for row, in oracle_cursor]
    oracle_db_conn.close()

    return wkfs


def get_workflow_from_db(configPath, queryCmd):
    '''
    get a list of :py:class:`WorkflowInfo` objects by parsing the oracle db
    indicated in ``config.yml`` pointed by configpath.

    :param str configPath: path of config file
    :param str queryCmd: SQL query command
    :returns: list of :py:class:`WorkflowInfo`

    :rtype: list
    '''

    wf_list = []

    config = get_yamlconfig(configPath)
    if not config:
        return wf_list

    wfs = get_workflowlist_from_db(config, queryCmd)
    if wfs:
        wf_list = [workflowinfo.WorkflowInfo(wf) for wf in wfs]

    return wf_list


def cleanup_shortlog(desc):
    """
    clean up given string by:

    1. remove any HTML tag ``<>``
    2. remove square brackets label ``[]``
    3. remove char ``\\``
    4. replace successive whitespace with a single one
    5. remove single quote/double quote

    :param str desc: a description string
    :returns: a cleaned description string

    :rtype: str
    """

    cleaned = re.compile(r'<.*?>').sub('', desc)
    cleaned = re.compile(r'\[.*?\]').sub('', cleaned)
    cleaned = cleaned.replace('\\', '')
    cleaned = re.compile('\s+').sub(' ', cleaned)
    cleaned = cleaned.replace('"', "'").replace("'", '')

    return cleaned


def short_errorlog(log,
                   buzzwords=['error', 'fail',
                              'exception', 'maxrss', 'timeout'],
                   ignorewords=['start', 'begin', 'end', 'above', 'below']):
    r"""
    pruned the lengthy error logs extracted from wmstats to a short message,
    with a logic of combination of ``buzzwords`` and ``ignorewords``.

    - First if ``log`` is short enouggh that does not contain a ``\n``, return it.
    - Else split the log with common delimiters to list, then clean up each entry with :py:func:`cleanup_shortlog`,
        - from begining, if a entry does not contain any word in ``ignorewords`` list, it shall need attention;
        - if a entry contains any word in buzzwords, it shall be buzzed;
        - if a buzzed entry contains less than 3 words, it shall be skipped.( not informative enough)

    - if anything in buzzed list, return a string concatenating all buzzed entries;
    - else if anything in attentioned list, return the first entry;
    - else returns the first entry after clean up only.

    :param str log: length log string from wmstats
    :param list buzzwords: list of words that shall draw attention
    :param list ignorewords: list of words that shall be ignored at any conditions
    :returns: shorted log that shall reflect key information

    :rtype: str
    """

    log = log.strip()
    if '\n' not in log:
        return log

    piecesList = re.split('; |, |:|\*|\n+', log)
    piecesList = [cleanup_shortlog(x) for x in piecesList]

    attentionedPieces = list()
    buzzedPieces = list()

    for piece in piecesList:
        piece = piece.strip()
        raw = piece.lower()

        if any(kw in raw for kw in ignorewords):
            continue
        attentionedPieces.append(piece)

        if any(kw in raw for kw in buzzwords):
            buzzedPieces.append(piece)

    if buzzedPieces:
        buzzedPieces = [x for x in set(buzzedPieces) if len(
            x.split(' ')) > 2]  # too short to be informative
    if buzzedPieces:
        return '; '.join(buzzedPieces)

    # should save exceptional error logs to enrich buzzwords
    if attentionedPieces:
        return attentionedPieces[0]
    else:
        return piecesList[0]


def extract_keywords(description,
                     buzzwords=['error', 'errors', 'errormsg', 'fail',
                                'failed', 'failure', 'kill', 'killed', 'exception'],
                     blacklistwords=['start', 'begin',
                                     'end', 'above', 'below'],
                     whitelistwords=['timeout', 'maxrss', 'nojobreport']):
    """
    extract keywords from shortened error log,
    with a logic of combination of ``buzzwords``, ``blacklistwords`` and ``whitelistwords``.

    For each word in the ``description``, if it's in ``whitelistwords``, add to return;
    if any word in ``buzzwords`` is a subset of this word, add to return.
    In the end, if any word in ``blacklistwords`` shows up in the to-return list, removes it.

    :param str description: shortened error log
    :param list buzzwords: list of words that shall draw attention
    :param list blacklistwords: list of words that should not be treated as keyword
    :param list whitelistwords: list of words that will always be treated as keyword
    :returns: a set of keywords

    :rtype: set
    """

    kwset = set()

    for word in re.compile('\w+').findall(description):
        word = word.strip()
        raw = word.lower()

        if any(kw in raw for kw in whitelistwords):
            kwset.add(word)

        for kw in buzzwords:
            if kw in raw and (raw not in buzzwords):
                kwset.add(word)

    for kw in blacklistwords:
        kwset.discard(kw)

    return kwset


def error_logs(workflow):
    """
    Given a :py:class:`WorkflowInfo`, builds up a structured entity
    representing all available necessary error information via WorkflowInfo's property.::

        {'taskName' :
            {'errorCode' :
                {'siteName' :

                    [
                        {
                            'secondaryErrorCodes' : [],
                            'errorKeywords' : [],
                            'errorChain' : [(typecode, shortLog), ...]
                        },

                        ...
                    ]
                }
            }
        }

    :param workflow: A :py:class:`WorkflowInfo` object
    :returns: error info parsed from logs

    :rtype: collections.defaultdict
    """

    error_logs = defaultdict(
        lambda: defaultdict(
            lambda: defaultdict(
                lambda: []
            )
        )
    )

    wf_jobdetail = workflow._get_jobdetail()
    wf_stepinfo = wf_jobdetail['result'][0].get(workflow.workflow, {})

    if not wf_stepinfo:
        return error_logs

    for stepname, stepdata in wf_stepinfo.items():
        _taskName = stepname.split('/')[-1]
        # Get the errors from both 'jobfailed' and 'submitfailed' details
        for error, sitedata in [
            (error, sitedata) for status in ['jobfailed', 'submitfailed']
            for error, sitedata in stepdata.get(status, {}).items()
        ]:
            if error == '0':
                continue
            _errorcode = int(error)

            for _sitename, siteinfo in sitedata.items():
                _errorsamples = list()

                for sample in siteinfo['samples']:
                    _timestamp = sample['timestamp']
                    errorcells = [e for cateInfo in sample['errors'].values()
                                  for e in cateInfo]
                    errorcells_unique = list()
                    for ec in errorcells:
                        if ec in errorcells_unique:
                            continue
                        errorcells_unique.append(ec)

                    _secondaryCodes = list()
                    _errorKeywords = list()
                    _errorChainAsDicts = list()

                    for ec in errorcells_unique:

                        type_ = ec['type']
                        code_ = ec['exitCode']
                        shortdetail_ = short_errorlog(ec['details'])

                        if code_ != _errorcode:
                            _secondaryCodes.append(code_)

                        _errorKeywords.extend(list(
                            extract_keywords(' '.join([
                                type_,
                                shortdetail_
                            ]))
                        ))

                        _errorChainAsDicts.append({
                            "errorType": type_,
                            "exitCode": code_,
                            "description": shortdetail_
                        })

                    _errorsamples.append({
                        'secondaryErrorCodes': list(set(_secondaryCodes)),
                        'errorKeywords': list(set(_errorKeywords)),
                        'errorChain': _errorChainAsDicts,
                        'timeStamp': _timestamp
                    })

                error_logs[_taskName][_errorcode][_sitename] = _errorsamples

    return error_logs


def error_summary(workflow):
    """
    Given a :py:class:`WorkflowInfo`,  build a minimal error summary via :py:func:`workflow.get_errors` method.::

        {'taskName':
            {'errors': [
                {
                    'errorCode' : errorCode,
                    'siteName' : siteName,
                    'counts' : counts
                },

                 ...

                ],
            'siteNotReported': []
            }
        }

    :param workflow: A :py:class:`WorkflowInfo` object
    :returns: A dict representing mimnial error summary

    :rtype: dict
    """

    error_summary = dict()

    errorInfo = workflow.get_errors(get_unreported=True)
    if not errorInfo:
        return error_summary

    for fullTaskName, taskErrors in errorInfo.items():
        taskName = fullTaskName.split('/')[-1]
        if not taskName:
            continue

        errorList = list()
        noReportSite = list(taskErrors.get('NotReported', {}).keys())
        for errorCode, siteCnt in taskErrors.items():
            if errorCode == 'NotReported':
                continue

            for siteName, counts in siteCnt.items():
                errorList.append({
                    'errorCode': int(errorCode),
                    'siteName': siteName,
                    'counts': counts
                })

        error_summary[taskName] = {
            'errors': errorList,
            'siteNotReported': noReportSite
        }

    return error_summary


def populate_error_for_workflow(workflow):
    """
    Given a :py:class:`WorkflowInfo`,  build an ultimate error summary with
    :py:func:`error_logs`, :py:func:`error_summary` method and wmstats API.

    :param workflow: A :py:class:`WorkflowInfo` object
    :returns: A dict representing all available error info

    :rtype: dict
    """

    if isinstance(workflow, str):
        workflow = workflowinfo.WorkflowInfo(workflow)
    assert(isinstance(workflow, workflowinfo.WorkflowInfo))

    workflow_summary = {
        "name": workflow.workflow,
        "status": None,
        "type": None,
        "failureRate": 0.,
        "totalError": 0,
        "failureKeywords": [],
        "transitions": [],
        "tasks": {}
    }

    workflow_summary['failureRate'] = workflow.get_failure_rate()
    wf_reqdetail = workflow._get_reqdetail()
    #save_json(wf_reqdetail, 'wf_reqdetail')

    wfData = wf_reqdetail.get(workflow.workflow, {})
    if not wfData:
        return workflow_summary

    # info extracted from wmstats request API
    agentJobInfo = wfData.get('AgentJobInfo', {})
    requestStatus = wfData.get('RequestStatus', None)
    requestType = wfData.get('RequestType', None)
    if not all([agentJobInfo, requestStatus, requestType]):
        return workflow_summary

    requestTransition = wfData.get('RequestTransition', [])

    workflow_summary['status'] = requestStatus
    workflow_summary['type'] = requestType
    workflow_summary['transitions'] = requestTransition

    nfailure = 0
    for agent, agentdata in agentJobInfo.items():
        status = agentdata.get('status', {})
        tasks = agentdata.get('tasks', {})
        if not all([status, tasks]):
            continue

        for ftype, num in status.get('failure', {}).items():
            nfailure += num

        for taskFullName, taskData in tasks.items():
            taskName = taskFullName.split('/')[-1]

            inputTask = None
            if len(taskFullName.split('/')) > 3:
                inputTask = taskFullName.split('/')[-2]

            jobType = taskData.get('jobtype', None)
            taskStatus = taskData.get('status', {})

            taskSiteError = dict()

            if taskStatus and taskStatus.get('failure', {}):
                for site, siteData in taskData.get('sites', {}).items():
                    errCnt = 0
                    errCnts = siteData.get('failure', {})
                    if not errCnts:
                        continue

                    for ftype, cnt in errCnts.items():
                        errCnt += cnt

                    taskSiteError[site] = errCnt

            _task = workflow_summary['tasks'].get(taskName, None)
            if _task:
                if 'jobType' not in _task.keys():
                    _task["jobType"] = jobType
                if 'inputTask' not in _task.keys():
                    _task['inputTask'] = inputTask
                if 'siteErrors' not in _task.keys():
                    _task["siteErrors"] = taskSiteError
                else:
                    for site, errors in taskSiteError.items():
                        if site in _task["siteErrors"].keys():
                            _task["siteErrors"][site] += errors
                        else:
                            _task["siteErrors"][site] = errors

                workflow_summary['tasks'][taskName] = _task
            else:
                workflow_summary['tasks'][taskName] = {
                    "inputTask": inputTask,
                    "jobType": jobType,
                    "siteErrors": taskSiteError,
                    "errors": [],
                    "siteNotReported": []
                }

    # remove tasks that does not have any error
    taskToDel = list()
    for taskname, taskinfo in workflow_summary['tasks'].items():
        if 'siteErrors' in taskinfo and (not taskinfo['siteErrors']):
            taskToDel.append(taskname)
    for taskname in taskToDel:
        workflow_summary['tasks'].pop(taskname, None)

    workflow_summary['totalError'] = nfailure

    if workflow_summary['status'] != 'rejected':

        wf_errorSummary = error_summary(workflow)
        wf_errorLog = error_logs(workflow)

        # add information from errorSummary
        for taskName, taskErrors in wf_errorSummary.items():
            if taskName in workflow_summary['tasks'].keys():
                workflow_summary['tasks'][taskName].update(taskErrors)

        # add information from errorLog
        for taskName, taskErrorLogInfo in wf_errorLog.items():

            if taskName not in workflow_summary['tasks'].keys():
                continue
            for errorCode, siteInfo in taskErrorLogInfo.items():
                for site, info in siteInfo.items():

                    for e in workflow_summary['tasks'][taskName].get('errors', []):
                        if e.get('siteName', None) != site:
                            continue
                        if e.get('errorCode', None) != errorCode:
                            continue

                        if len(info):
                            e.update(info[0])

        # fill failureKeywords list
        allKeywords = [kw
                       for task in workflow_summary['tasks'].values()
                       for error in task.get('errors', [])
                       for kw in error.get('errorKeywords', [])
                       ]
        workflow_summary['failureKeywords'] = list(set(allKeywords))

    # last step, nest in task key(TaskName) as a key-value pair
    tasksAsList = []
    for taskname, taskinfo in workflow_summary['tasks'].items():
        taskinfo.update({"name": taskname})
        taskinfo['siteErrors'] = [
            {
                "site": site,
                "counts": counts
            } for site, counts in taskinfo['siteErrors'].items()
        ]  # convert 'siteErrors' from a dict to a list of dict
        tasksAsList.append(taskinfo)
    workflow_summary['tasks'] = tasksAsList

    return workflow_summary


def _get_acdc_response(wfstr):
    """
    debug
    """

    from cmstoolbox.webtools import get_json
    response = get_json(
        'cmsweb.cern.ch',
        '/couchdb/acdcserver/_design/ACDC/_view/byCollectionName',
        {'key': '"{0}"'.format(
                wfstr), 'include_docs': 'true', 'reduce': 'false'},
        use_cert=True
    )
    return response


def filter_n_collector(res, q, minFailureRate=0.2):

    while not q.empty():
        wf = q.get()
        try:
            if wf.get_failure_rate() > minFailureRate:
                res.append(populate_error_for_workflow(wf))
        except:
            pass
        q.task_done()
    return True


def main():

    start_time = time.time()

    # status_location = '/home/wsi/workdir/statuses.json' # dummy
    CONFIG_FILE_PATH = os.path.join(os.path.dirname(
        os.path.abspath(__file__)), 'config.yml')
    #DB_QUERY_CMD = "SELECT NAME FROM CMS_UNIFIED_ADMIN.WORKFLOW WHERE WM_STATUS LIKE 'running%'"
    DB_QUERY_CMD = "SELECT NAME FROM CMS_UNIFIED_ADMIN.workflow WHERE lower(STATUS) LIKE '%manual%'"
    wfs = get_workflow_from_db(CONFIG_FILE_PATH, DB_QUERY_CMD)

    print("Number of workflows retrieved from Oracle DB: ", len(wfs))
    invalidate_caches()

    try:
        from Queue import Queue
    except ImportError:
        from queue import Queue # pylint: disable=import-error
    q = Queue()
    num_threads = min(150, len(wfs))

    for wf in wfs:
        q.put(wf)
    results = list()
    for _ in range(num_threads):
        t = threading.Thread(target=filter_n_collector, args=(results, q, ))
        t.daemon = True  # seting threads as "daemon" allows main program to exit
        # eventually even if these dont finish correctly.
        t.start()
    q.join()
    print("Number of workflows that has >20% failure rate: ", len(results))

    elasped_time = time.time() - start_time
    print('Takes {0} s in total.'.format(elasped_time))

    ###################################

    # key, ws = wfs[1]
    # print(key)
    # pprint(populate_error_for_workflow(ws))
    # pprint(error_summary(ws))
    # pprint(error_logs(ws))
    # save_json(error_logs(ws), 'wf_errorlog')
    # save_json(populate_error_for_workflow(ws), 'wf_document')
    save_json(results, 'wfc_results')


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import os
import time

import schedule
import workflowmonit.sendToMonit as wms

CRED_FILE_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'credential.yml')
CONFIG_FILE_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'config.yml')
LOGDIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Logs')
LOGGING_CONFIG = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'configLogging.yml')

schedule.every(30).minutes.do(wms.main)


if __name__ == "__main__":

    while True:
        schedule.run_pending()
        time.sleep(1)



"""
Generates Mako templates
"""

import os

import mako.lookup

from .. import serverconfig

TEMPLATES_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                             'templates')

def render(template, **kwargs):
    """
    Function to generate mako template

    :param str template: Name of the mako template to generate
    :param kwargs: The keywords to pass to the rendered template
    :returns: Rendered webpage from the selected template
    :rtype: str
    """

    return mako.lookup.TemplateLookup(
        directories=[TEMPLATES_DIR],
        module_directory=os.path.join(
            serverconfig.config_dict()['workspace'], 'mako_modules')
        ).get_template(template).render(**kwargs)



#!/user/bin/env python
from __future__ import print_function
import os
import sys
import sqlite3

import workflowmonit.workflowCollector as wc

CONFIG_FILE_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'config.yml')

def main():

    config = wc.get_yamlconfig(CONFIG_FILE_PATH)
    if not config:
        sys.exit('Config file: {} not exist, exiting..'.format(CONFIG_FILE_PATH))
    dbPath = config.get(
        'workflow_status_db',
        os.path.join(os.path.dirname(os.path.abspath(__file__)), 'workflow_status.sqlite')
        )

    conn = sqlite3.connect(dbPath)
    with conn:
        c = conn.cursor()
        for row in c.execute("SELECT * FROM workflowStatuses WHERE status LIKE '%archived' ORDER BY failurerate"):
            print('[ {1:^15} ]\t{2:.6f}\t{0}'.format(*row))

if __name__ == "__main__":
    main()
